from typing import override

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QWidget,
    QPushButton,
    QLineEdit,
)

from caqtus.types.expression import Expression
from .._delegate import TimeLaneDelegate


class DigitalTimeLaneDelegate(TimeLaneDelegate):
    @override
    def createEditor(self, parent, option, index) -> QWidget:
        cell_value = index.data(Qt.ItemDataRole.EditRole)
        if isinstance(cell_value, bool):
            return CheckedButton(parent)
        elif isinstance(cell_value, Expression):
            return QLineEdit(parent)
        else:
            raise TypeError(f"Invalid type for value: {type(cell_value)}")

    @override
    def setEditorData(self, editor, index):
        cell_value = index.data(Qt.ItemDataRole.EditRole)
        if isinstance(cell_value, bool):
            assert isinstance(editor, CheckedButton)
            # We invert the value saved in the model because then when the user open
            # the editor, the button will already have changed its state, and they
            # don't have to click it a second time to change the value.
            editor.setChecked(not cell_value)
        elif isinstance(cell_value, Expression):
            assert isinstance(editor, QLineEdit)
            editor.setText(str(cell_value))
        else:
            raise TypeError(f"Invalid type for value: {type(cell_value)}")

    @override
    def setModelData(self, editor, model, index):
        cell_value = index.data(Qt.ItemDataRole.EditRole)
        if isinstance(cell_value, bool):
            assert isinstance(editor, CheckedButton)
            model.setData(index, editor.isChecked(), Qt.ItemDataRole.EditRole)
        elif isinstance(cell_value, Expression):
            assert isinstance(editor, QLineEdit)
            model.setData(index, Expression(editor.text()), Qt.ItemDataRole.EditRole)
        else:
            raise TypeError(f"Invalid type for value: {type(cell_value)}")


class CheckedButton(QPushButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setCheckable(True)
        self.toggled.connect(self.on_toggled)

    @override
    def setChecked(self, a0: bool) -> None:  # type: ignore[reportIncompatibleMethodOverride]
        super().setChecked(a0)
        self.on_toggled(a0)

    def on_toggled(self, checked: bool):
        if checked:
            self.setText("Enabled")
        else:
            self.setText("Disabled")



from cattrs.gen import override

from ._json import (
    JSON,
    is_valid_json,
    is_valid_json_dict,
    is_valid_json_list,
    JsonDict,
    JsonList,
)
from .converters import (
    unstructure,
    converters,
    structure,
    register_unstructure_hook,
    register_structure_hook,
    to_json,
    from_json,
    copy_converter,
)
from .customize import customize
from .strategies import include_subclasses, include_type, configure_tagged_union

__all__ = [
    "converters",
    "structure",
    "unstructure",
    "register_structure_hook",
    "register_unstructure_hook",
    "to_json",
    "from_json",
    "customize",
    "override",
    "include_subclasses",
    "configure_tagged_union",
    "include_type",
    "JSON",
    "JsonDict",
    "JsonList",
    "is_valid_json",
    "is_valid_json_dict",
    "is_valid_json_list",
    "copy_converter",
]



import copy
from collections.abc import Sequence
from typing import Optional

from PySide6.QtWidgets import (
    QWidget,
    QGroupBox,
    QHBoxLayout,
    QTableWidget,
    QTableWidgetItem,
)

from caqtus.device.sequencer import ChannelConfiguration
from .channel_output_editor import ChannelOutputEditor


class SequencerChannelWidget[ChannelT: ChannelConfiguration](QWidget):
    """A widget that allow to edit the channel configurations of a sequencer.

    Attributes:
        channel_table: A table widget that shows the list of channels of the sequencer.
    """

    def __init__(self, channels: Sequence[ChannelT], parent: Optional[QWidget] = None):
        super().__init__(parent)

        # We use a table widget and not a list widget because we want to have headers
        self.channel_table = QTableWidget(len(channels), 1, self)
        self.channel_table.horizontalHeader().setStretchLastSection(True)
        self.channel_table.horizontalHeader().hide()
        self.group_box = QGroupBox(self)
        self.channel_output_editor: Optional[ChannelOutputEditor] = None
        self._populate_group_box()
        self.channels = tuple(channels)

        layout = QHBoxLayout(self)
        self.setLayout(layout)
        layout.addWidget(self.channel_table)
        layout.addWidget(self.group_box, 1)

        self._populate_channel_list()
        self.channel_table.currentItemChanged.connect(self._on_current_item_changed)
        self.channel_table.itemChanged.connect(self._on_item_changed)
        self.group_box.setVisible(False)

    def _populate_channel_list(self) -> None:
        self.channel_table.clear()
        for row, channel in enumerate(self.channels):
            item = QTableWidgetItem(channel.description)
            self.channel_table.setItem(row, 0, item)
        channel_labels = [self.channel_label(row) for row in range(len(self.channels))]
        self.channel_table.setVerticalHeaderLabels(channel_labels)

    def _populate_group_box(self) -> None:
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.group_box.setLayout(layout)

    def _on_current_item_changed(
        self, current: Optional[QTableWidgetItem], previous: Optional[QTableWidgetItem]
    ) -> None:
        if previous is not None:
            assert self.channel_output_editor is not None
            output = self.channel_output_editor.get_output()
            row = previous.row()
            self.channels[row].output = output
        self.set_preview_item(current)

    def _on_item_changed(self, item: QTableWidgetItem) -> None:
        if item is self.channel_table.currentItem():
            self.set_preview_item(item)

    def set_preview_item(self, item: Optional[QTableWidgetItem]):
        if item is not None:
            self.group_box.setVisible(True)
            self.group_box.setTitle(item.text())
            row = item.row()
            channel = self.channels[row]
            previous_editor = None
            if self.channel_output_editor is not None:
                previous_editor = self.channel_output_editor
            self.channel_output_editor = ChannelOutputEditor(channel.output, self)
            layout = self.group_box.layout()
            assert layout is not None
            layout.addWidget(self.channel_output_editor)
            if previous_editor:
                previous_editor.deleteLater()
        else:
            self.group_box.setVisible(False)

    def channel_label(self, row: int) -> str:
        return str(row)

    def get_channel_configurations(self) -> tuple[ChannelT, ...]:
        current_item = self.channel_table.currentItem()
        if current_item is not None:
            assert self.channel_output_editor is not None
            output = self.channel_output_editor.get_output()
            row = current_item.row()
            self.channels[row].output = output
        for row, channel in enumerate(self.channels):
            item = self.channel_table.item(row, 0)
            assert item is not None
            channel.description = item.text()
        return copy.deepcopy(self.channels)



"""
This module defines what interface the devices must satisfy to be used in an
experiment.
For each instrument to be used on the experiment, there must be a corresponding class
inheriting from :class:`Device` that contains the low-level logic to control the actual
instrument.

The only place where actual communication to an instrument can occur during an
experiment is inside the methods of a device.

Below is a very basic usage example of a device that satisfies this protocol:

.. code-block:: python

        with RotationState(name="stage", port="COM0") as rotation_stage:
            for angle in range(0, 360, 10):
                rotation_stage.update_parameters(angle=angle)
"""

from ._device import Device
from ._installed_devices_discovery import load_installed_devices
from .runtime_device import RuntimeDevice

__all__ = ["Device", "RuntimeDevice", "load_installed_devices"]



import abc
from typing import Protocol

from ._experiment_session import ExperimentSession
from .async_session import AsyncExperimentSession


class StorageManager(Protocol):
    """Used to create a new experiment sessions."""

    @abc.abstractmethod
    def __call__(self) -> ExperimentSession:
        """Create a new experiment session."""

        raise NotImplementedError

    def session(self) -> ExperimentSession:
        """Create a new experiment session."""

        return self()

    @abc.abstractmethod
    def async_session(self) -> AsyncExperimentSession:
        """Create a new asynchronous experiment session."""

        raise NotImplementedError


# Deprecated alias to StorageManager
ExperimentSessionMaker = StorageManager



from ._attrs import build_attrs_class_editor, AttributeOverride
from ._editor_builder import EditorBuilder, TypeNotRegisteredError, EditorFactory
from ._enum import generate_enum_editor
from ._literal import build_literal_editor

__all__ = [
    "EditorBuilder",
    "TypeNotRegisteredError",
    "build_attrs_class_editor",
    "build_literal_editor",
    "generate_enum_editor",
    "EditorFactory",
    "AttributeOverride",
]



from typing import Optional

from PySide6 import QtWidgets
from PySide6.QtCore import QSize, QModelIndex, Qt, QRectF
from PySide6.QtGui import QTextDocument, QAbstractTextDocumentLayout
from PySide6.QtWidgets import (
    QStyledItemDelegate,
    QStyleOptionViewItem,
    QWidget,
)


class HTMLItemDelegate(QStyledItemDelegate):
    """A delegate that can renders HTML when painting items in a view."""

    def __init__(self, parent: Optional[QWidget] = None):
        self.doc = QTextDocument()
        super().__init__(parent)

    def get_text_to_render(self, index: QModelIndex) -> str:
        """Return the text to render in the delegate.

        Override this method to customize the rendering of the text.
        """

        return str(index.data(role=Qt.DisplayRole))

    def paint(self, painter, option, index):
        options = QtWidgets.QStyleOptionViewItem(option)
        self.initStyleOption(options, index)
        painter.save()
        self.doc.setTextWidth(options.rect.width())
        text = self.get_text_to_render(index)
        self.doc.setHtml(text)
        self.doc.setDefaultFont(options.font)
        options.text = ""
        options.widget.style().drawControl(
            QtWidgets.QStyle.ControlElement.CE_ItemViewItem, options, painter
        )
        painter.translate(options.rect.left(), options.rect.top())
        clip = QRectF(0, 0, options.rect.width(), options.rect.height())
        painter.setClipRect(clip)
        ctx = QAbstractTextDocumentLayout.PaintContext()
        ctx.clip = clip
        self.doc.documentLayout().draw(painter, ctx)
        painter.restore()

    def sizeHint(self, option, index):
        options = QStyleOptionViewItem(option)
        self.initStyleOption(options, index)
        doc = QTextDocument()
        doc.setHtml(options.text)
        doc.setTextWidth(options.rect.width())
        return QSize(doc.idealWidth(), doc.size().height())



"""Contains components to edit a digital time lane."""

from ._delegate import DigitalTimeLaneDelegate
from ._model import DigitalTimeLaneModel

__all__ = ["DigitalTimeLaneModel", "DigitalTimeLaneDelegate"]



from __future__ import annotations

import attrs

from ._path import PureSequencePath


@attrs.frozen
class ShotId:
    """Unique identifier for a shot in a sequence."""

    sequence_path: PureSequencePath
    index: int



import contextlib
from collections.abc import Callable, Iterator
from typing import (
    Self,
    ParamSpec,
    TypeVar,
    LiteralString,
    Any,
    final,
    AsyncIterator,
)

from ._async_converter import AsyncConverter
from .rpc import Proxy
from ..runtime import Device
from ...utils.context_managers import aclose_on_error

T = TypeVar("T")
P = ParamSpec("P")


class DeviceProxy[DeviceType: Device]:
    """Proxy to a remote device.

    This class is used on the client side to interact with a device running on a remote
    server.
    It provides asynchronous methods to get attributes and call methods remotely
    without blocking the client.
    """

    @final
    def __init__(
        self,
        async_converter: AsyncConverter,
        device_type: Callable[P, DeviceType],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        self.async_converter = async_converter
        self._device_type = device_type
        self._args = args
        self._kwargs = kwargs
        self._device_proxy: Proxy[DeviceType]

        self._async_exit_stack = contextlib.AsyncExitStack()

    async def __aenter__(self) -> Self:
        async with aclose_on_error(self._async_exit_stack):
            self._device_proxy = await self._async_exit_stack.enter_async_context(
                self.async_converter.call_proxy_result(
                    self._device_type, *self._args, **self._kwargs
                )
            )
            await self._async_exit_stack.enter_async_context(
                self.async_context_manager(self._device_proxy)
            )
        return self

    async def get_attribute(self, attribute_name: LiteralString) -> Any:
        return await self.async_converter.get_attribute(
            self._device_proxy, attribute_name
        )

    async def call_method(
        self,
        method_name: LiteralString,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        return await self.async_converter.call_method(
            self._device_proxy, method_name, *args, **kwargs
        )

    def call_method_proxy_result(
        self,
        method_name: LiteralString,
        *args: Any,
        **kwargs: Any,
    ) -> contextlib.AbstractAsyncContextManager[Proxy]:
        return self.async_converter.call_method_proxy_result(
            self._device_proxy, method_name, *args, **kwargs
        )

    def async_context_manager(
        self, proxy: Proxy[contextlib.AbstractContextManager[T]]
    ) -> contextlib.AbstractAsyncContextManager[Proxy[T]]:
        return self.async_converter.async_context_manager(proxy)

    def async_iterator(self, proxy: Proxy[Iterator[T]]) -> AsyncIterator[T]:
        return self.async_converter.async_iterator(proxy)

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        await self._async_exit_stack.__aexit__(exc_type, exc_value, traceback)



"""Defines user expressions that can be evaluated later."""

from ._expression import Expression, expression_builtins, DEFAULT_BUILTINS

__all__ = ["Expression", "expression_builtins", "DEFAULT_BUILTINS"]



from typing import Optional, TYPE_CHECKING

from ._instructions import TimedInstruction
from ._to_time_array import convert_to_change_arrays

if TYPE_CHECKING:
    from matplotlib.axes import Axes


def plot_instruction(
    instruction: TimedInstruction, ax: Optional["Axes"] = None
) -> "Axes":
    """Plot the instruction on the given axis."""

    import matplotlib.pyplot as plt

    if ax is None:
        fig, ax = plt.subplots()

    change_times, change_values = convert_to_change_arrays(instruction)

    ax.step(change_times, change_values, where="post")
    ax.set_xlabel("Time [ticks]")
    ax.set_ylabel("Value")
    return ax



from abc import ABC, abstractmethod
from typing import Iterable

import attrs
import numpy as np

from .._image_type import Image, is_image
from .._shape import Width, Height


@attrs.define
class ROI(ABC):
    """Base class for regions of interest inside an image.

    Attributes:
        original_image_size: The size of the original image as (width, height).
    """

    original_image_size: tuple[Width, Height] = attrs.field(
        on_setattr=attrs.setters.validate,
    )

    @original_image_size.validator  # type: ignore
    def _validate_original_image_size(self, _, value):
        if not isinstance(value, tuple):
            raise TypeError(f"original_image_size must be a tuple, got {type(value)}")
        if len(value) != 2:
            raise ValueError(
                "original_image_size must be a tuple (width, height) with two elements"
            )
        if not all(isinstance(x, int) for x in value):
            raise ValueError("original_image_size must be a tuple of integers")
        if not all(x > 0 for x in value):
            raise ValueError("original_image_size must be a tuple of positive integers")

    def get_mask(self) -> np.ndarray:
        """A boolean array with the same shape as the original image.

        True values indicate that the pixel is part of the region of interest."""

        mask = np.full(self.original_image_size, False)
        mask[*self.get_indices()] = True
        return mask

    @abstractmethod
    def get_indices(self) -> tuple[Iterable[int], Iterable[int]]:
        """Return the indices of the pixels in the ROI."""

        raise NotImplementedError

    @property
    def original_width(self) -> int:
        """Return the width of the original image."""

        return self.original_image_size[0]

    @property
    def original_height(self) -> int:
        """Return the height of the original image."""

        return self.original_image_size[1]

    def apply[
        T: np.generic
    ](self, image: Image[T]) -> np.ma.MaskedArray[tuple[Width, Height], np.dtype[T]]:
        """Apply the ROI to an image.

        Returns:
            A masked array with the pixels outside the ROI masked.
        """

        if not is_image(image):
            raise TypeError(
                f"Image must be a numpy array with two dimensions, got {image}"
            )

        if image.shape != self.original_image_size:
            raise ValueError(
                f"Image shape {image.shape} does not match the roi original image size "
                f"{self.original_image_size}"
            )

        return np.ma.MaskedArray(image, mask=~self.get_mask())



from __future__ import annotations

from collections.abc import Callable
from typing import Never, overload, Any, TypeVar, Generic

import attrs
from typing_extensions import TypeIs

T = TypeVar("T", covariant=True)
R = TypeVar("R", covariant=True)
E = TypeVar("E", covariant=True)


@attrs.frozen(repr=False, str=False)
class Success(Generic[T]):
    """A successful result containing a value of type T."""

    value: T

    def map(self, func: Callable[[T], R]) -> Success[R]:
        return Success(func(self.value))

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> str:
        return f"Success({self.value!r})"

    def result(self) -> T:
        return self.value

    def content(self) -> T:
        """Return the wrapped successful value."""

        return self.value


def is_success(result: Success[T] | Failure[Any]) -> TypeIs[Success[T]]:
    """Check if a result is a success."""

    return isinstance(result, Success)


def is_failure(result: Any) -> TypeIs[Failure]:
    """Check if a result is a failure."""

    return isinstance(result, Failure)


def is_failure_type(
    result: Any, error_type: tuple[type[E], ...] | type[E]
) -> TypeIs[Failure[E]]:
    """Check if a result is a failure and contains a specific error type."""

    return is_failure(result) and isinstance(result._error, error_type)


@attrs.frozen(repr=False, str=False)
class Failure(Generic[E]):
    """A failed result containing an error code of type E."""

    _error: E

    def map(self, func: Callable) -> Failure[E]:
        return self

    def __str__(self) -> str:
        return str(self._error)

    def exception[Exc: BaseException](self: Failure[Exc]) -> Exc:
        raise self._error


@overload
def unwrap(value: Success[T]) -> T: ...


@overload
def unwrap(value: Failure[Exception]) -> Never: ...


def unwrap(value):
    """Unwrap a result when the failure case is an exception.

    This function can be used to recover the wrapped value from a Success or raise the
    wrapped exception from a Failure.

    Args:
        value: The result to unwrap.
            If the value is a Failure, its content must be an exception.

    Returns:
        The value wrapped if the argument passes is a Success.

    Raises:
        The exception wrapped if the argument passed is a Failure containing an
        exception.
    """

    if isinstance(value, Success):
        return value.value
    else:
        raise value._error



"""Contains utility functions for defining and discretizing times."""

from ._timing import (
    Time,
    to_time,
    to_time_bounds,
    get_step_bounds,
    start_tick,
    stop_tick,
    number_ticks,
    ps,
)

__all__ = [
    "Time",
    "to_time",
    "to_time_bounds",
    "get_step_bounds",
    "start_tick",
    "stop_tick",
    "number_ticks",
    "ps",
]



#!/usr/bin/python
from PySide6 import QtCore, QtWidgets

from ..constants import Z_VAL_NODE, ITEM_CACHE_MODE, LayoutDirectionEnum, NodeEnum


class AbstractNodeItem(QtWidgets.QGraphicsItem):
    """
    The base class of all node qgraphics item.
    """

    def __init__(self, name="node", parent=None):
        super(AbstractNodeItem, self).__init__(parent)
        self.setFlags(
            QtWidgets.QGraphicsItem.ItemIsSelectable
            | QtWidgets.QGraphicsItem.ItemIsMovable
        )
        self.setCacheMode(ITEM_CACHE_MODE)
        self.setZValue(Z_VAL_NODE)
        self._properties = {
            "id": None,
            "name": name.strip(),
            "color": (13, 18, 23, 255),
            "border_color": (46, 57, 66, 255),
            "text_color": (255, 255, 255, 180),
            "type_": "AbstractBaseNode",
            "selected": False,
            "disabled": False,
            "visible": False,
            "layout_direction": LayoutDirectionEnum.HORIZONTAL.value,
        }
        self._width = NodeEnum.WIDTH.value
        self._height = NodeEnum.HEIGHT.value

    def __repr__(self):
        return "{}.{}('{}')".format(self.__module__, self.__class__.__name__, self.name)

    def boundingRect(self):
        return QtCore.QRectF(0.0, 0.0, self._width, self._height)

    def mousePressEvent(self, event):
        """
        Re-implemented to update "self._properties['selected']" attribute.

        Args:
            event (QtWidgets.QGraphicsSceneMouseEvent): mouse event.
        """
        self._properties["selected"] = True
        super(AbstractNodeItem, self).mousePressEvent(event)

    def setSelected(self, selected):
        self._properties["selected"] = selected
        super(AbstractNodeItem, self).setSelected(selected)

    def draw_node(self):
        """
        Re-draw the node item in the scene with proper
        calculated size and widgets aligned.

        (this is called from the builtin custom widgets.)
        """
        return

    def pre_init(self, viewer, pos=None):
        """
        Called before node has been added into the scene.

        Args:
            viewer (NodeGraphQt.widgets.viewer.NodeViewer): main viewer.
            pos (tuple): the cursor pos if node is called with tab search.
        """
        return

    def post_init(self, viewer, pos=None):
        """
        Called after node has been added into the scene.

        Args:
            viewer (NodeGraphQt.widgets.viewer.NodeViewer): main viewer
            pos (tuple): the cursor pos if node is called with tab search.
        """
        return

    @property
    def id(self):
        return self._properties["id"]

    @id.setter
    def id(self, unique_id=""):
        self._properties["id"] = unique_id

    @property
    def type_(self):
        return self._properties["type_"]

    @type_.setter
    def type_(self, node_type="NODE"):
        self._properties["type_"] = node_type

    @property
    def layout_direction(self):
        return self._properties["layout_direction"]

    @layout_direction.setter
    def layout_direction(self, value=0):
        self._properties["layout_direction"] = value

    @property
    def size(self):
        return self._width, self._height

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, width=0.0):
        self._width = width

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, height=0.0):
        self._height = height

    @property
    def color(self):
        return self._properties["color"]

    @color.setter
    def color(self, color=(0, 0, 0, 255)):
        self._properties["color"] = color

    @property
    def text_color(self):
        return self._properties["text_color"]

    @text_color.setter
    def text_color(self, color=(100, 100, 100, 255)):
        self._properties["text_color"] = color

    @property
    def border_color(self):
        return self._properties["border_color"]

    @border_color.setter
    def border_color(self, color=(0, 0, 0, 255)):
        self._properties["border_color"] = color

    @property
    def disabled(self):
        return self._properties["disabled"]

    @disabled.setter
    def disabled(self, state=False):
        self._properties["disabled"] = state

    @property
    def selected(self):
        if self._properties["selected"] != self.isSelected():
            self._properties["selected"] = self.isSelected()
        return self._properties["selected"]

    @selected.setter
    def selected(self, selected=False):
        self.setSelected(selected)

    @property
    def visible(self):
        return self._properties["visible"]

    @visible.setter
    def visible(self, visible=False):
        self._properties["visible"] = visible
        self.setVisible(visible)

    @property
    def xy_pos(self):
        """
        return the item scene postion.
        ("node.pos" conflicted with "QGraphicsItem.pos()"
        so it was refactored to "xy_pos".)

        Returns:
            list[float]: x, y scene position.
        """
        return [float(self.scenePos().x()), float(self.scenePos().y())]

    @xy_pos.setter
    def xy_pos(self, pos=None):
        """
        set the item scene postion.
        ("node.pos" conflicted with "QGraphicsItem.pos()"
        so it was refactored to "xy_pos".)

        Args:
            pos (list[float]): x, y scene position.
        """
        pos = pos or [0.0, 0.0]
        self.setPos(pos[0], pos[1])

    @property
    def name(self):
        return self._properties["name"]

    @name.setter
    def name(self, name=""):
        self._properties["name"] = name
        self.setToolTip("node: {}".format(name))

    @property
    def properties(self):
        """
        return the node view attributes.

        Returns:
            dict: {property_name: property_value}
        """
        props = {"width": self.width, "height": self.height, "pos": self.xy_pos}
        props.update(self._properties)
        return props

    def viewer(self):
        """
        return the main viewer.

        Returns:
            NodeGraphQt.widgets.viewer.NodeViewer: viewer object.
        """
        if self.scene():
            return self.scene().viewer()

    def delete(self):
        """
        remove node view from the scene.
        """
        if self.scene():
            self.scene().removeItem(self)

    def from_dict(self, node_dict):
        """
        set the node view attributes from the dictionary.

        Args:
            node_dict (dict): serialized node dict.
        """
        node_attrs = list(self._properties.keys()) + ["width", "height", "pos"]
        for name, value in node_dict.items():
            if name in node_attrs:
                # "node.pos" conflicted with "QGraphicsItem.pos()"
                # so it's refactored to "xy_pos".
                if name == "pos":
                    name = "xy_pos"
                setattr(self, name, value)



from __future__ import annotations

import datetime
from typing import Optional, TYPE_CHECKING

from sqlalchemy import ForeignKey, DateTime, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from ._table_base import Base

if TYPE_CHECKING:
    from ._sequence_table import SQLSequence


class SQLSequencePath(Base):
    __tablename__ = "path"

    id_: Mapped[int] = mapped_column(name="id", primary_key=True, index=True)
    path: Mapped[str] = mapped_column(String(255), index=True, unique=True)
    parent_id: Mapped[Optional[int]] = mapped_column(
        ForeignKey("path.id", ondelete="CASCADE"), index=True
    )

    children: Mapped[list[SQLSequencePath]] = relationship(
        back_populates="parent", cascade="delete, delete-orphan", passive_deletes=True
    )
    parent: Mapped[SQLSequencePath] = relationship(
        back_populates="children", remote_side=[id_]
    )

    # Stored as timezone naive datetime, with the assumption that the timezone is UTC.
    creation_date: Mapped[datetime.datetime] = mapped_column(DateTime(timezone=False))
    sequence: Mapped[Optional["SQLSequence"]] = relationship(
        back_populates="path", cascade="all, delete", passive_deletes=True
    )

    def __str__(self):
        return str(self.path)



from collections.abc import Callable
from typing import Optional, NewType, TypedDict

from caqtus.utils.serialization import (
    JSON,
    is_valid_json_dict,
    JsonDict,
    copy_converter,
)
from ._protocol import TimeLaneSerializerProtocol
from ..timelane import TimeLane, TimeLanes
from ...expression import Expression

Tag = NewType("Tag", str)


type LaneDumper[L: TimeLane] = Callable[[L], JsonDict]
"""A function that serializes a time lane to a JSON dictionary."""

type LaneLoader[L: TimeLane] = Callable[[JsonDict], L]
"""A function that constructs a time lane from a JSON dictionary."""

converter = copy_converter()


class TimeLaneSerializer(TimeLaneSerializerProtocol):
    def __init__(self):
        self.dumpers: dict[type, tuple[LaneDumper, Tag]] = {}
        self.loaders: dict[Tag, LaneLoader] = {}

    def register_time_lane[
        L: TimeLane
    ](
        self,
        lane_type: type[L],
        dumper: LaneDumper[L],
        loader: LaneLoader[L],
        type_tag: Optional[str] = None,
    ) -> None:
        if type_tag is None:
            tag = Tag(lane_type.__qualname__)
        else:
            tag = Tag(type_tag)
        self.dumpers[lane_type] = (dumper, tag)
        self.loaders[tag] = loader

    def dump(self, lane: TimeLane) -> JsonDict:
        dumper, tag = self.dumpers[type(lane)]
        content = dumper(lane)
        if "type" in content:
            raise ValueError("The content already has a type tag.")
        content["type"] = tag
        return content

    def load(self, data: JsonDict) -> TimeLane:
        tag = data["type"]
        if not isinstance(tag, str):
            raise ValueError("Invalid type tag.")
        else:
            tag = Tag(tag)
        try:
            loader = self.loaders[tag]
        except KeyError:
            error = ValueError(f"Don't know how to load lane stored with tag `{tag}`.")
            error.add_note(
                "Make sure that the corresponding lane extension is registered."
            )
            raise error from None
        return loader(data)

    def unstructure_time_lanes(self, time_lanes: TimeLanes) -> JsonDict:
        return dict(
            step_names=converter.unstructure(time_lanes.step_names, list[str]),
            step_durations=converter.unstructure(
                time_lanes.step_durations, list[Expression]
            ),
            lanes={
                lane: self.dump(time_lane)
                for lane, time_lane in time_lanes.lanes.items()
            },
        )

    def structure_time_lanes(self, content: JsonDict) -> TimeLanes:
        step_names = converter.structure(content["step_names"], list[str])
        step_durations = converter.structure(
            content["step_durations"], list[Expression]
        )
        lanes_content = content["lanes"]
        if not is_valid_json_dict(lanes_content):
            raise ValueError("Invalid lanes content.")
        lanes = {}
        for lane, lane_content in lanes_content.items():
            if not is_valid_json_dict(lane_content):
                raise ValueError(f"Invalid content for lane {lane}.")
            lanes[lane] = self.load(lane_content)

        return TimeLanes(
            step_names=step_names,
            step_durations=step_durations,
            lanes=lanes,
        )


class TimeLanesDump(TypedDict):
    step_names: list[str]
    step_durations: list[Expression]
    lanes: dict[str, JSON]


def default_dumper(lane) -> JSON:
    raise NotImplementedError(f"Unsupported type {type(lane)}")



import copy
import functools
from typing import Optional, Protocol

import attrs
from PySide6.QtWidgets import QLineEdit, QVBoxLayout
from typing_extensions import deprecated

from caqtus.device import DeviceConfiguration
from caqtus.device.configuration import DeviceServerName
from caqtus.device.output_transform import EvaluableOutput
from caqtus.gui.condetrol.device_configuration_editors import DeviceConfigurationEditor
from caqtus.gui.condetrol.device_configuration_editors.camera_configuration_editor import (  # noqa E501
    RectangularROIEditor as RectangularROIWidget,
)
from caqtus.types.expression import Expression
from caqtus.types.image.roi import RectangularROI
from ._editor_builder import (
    EditorBuilder,
    EditorFactory,
    AttributeOverride,
    build_attrs_class_editor,
)
from ._expression_editor import ExpressionEditor
from ._output_transform_editor import OutputTransformEditor
from ._value_editor import ValueEditor


class GeneratedConfigEditor[C: DeviceConfiguration](DeviceConfigurationEditor[C]):
    def __init__(
        self,
        config: C,
        *,
        editor_factory: EditorFactory[C],
    ) -> None:
        super().__init__()
        self._editor = editor_factory()
        self._editor.set_value(config)
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._editor.widget)
        self.setLayout(layout)

    # TODO: Understand why need to silence pyright
    def get_configuration(self) -> C:  # type: ignore[reportIncompatibleMethodOverride]
        return self._editor.read_value()

    def set_editable(self, editable: bool) -> None:
        self._editor.set_editable(editable)


class DeviceConfigEditorFactory[C: DeviceConfiguration](Protocol):
    def __call__(self, config: C) -> GeneratedConfigEditor[C]: ...


_builder = EditorBuilder()


def generate_device_configuration_editor[
    C: DeviceConfiguration
](
    config_type: type[C],
    generator: EditorBuilder = _builder,
    **overrides: AttributeOverride,
) -> DeviceConfigEditorFactory[C]:
    """Generate a device configuration editor for the given configuration type.

    For more documentation, see :func:`~caqtus.gui.autogen.build_attrs_class_editor`.

    Args:
        config_type: The type of configuration to construct the editor for.
            Must be an attrs class.
        generator: Used to generate editors for the fields of the configuration.

    Returns:
        An automatically generated class of type
        :class:`~caqtus.gui.condetrol.device_configuration_editors.DeviceConfigurationEditor`
        that can be used to edit configurations with type `config_type`.
    """

    if not attrs.has(config_type):
        raise TypeError("config_type must be an attrs class")

    config_editor_factory = build_attrs_class_editor(
        config_type, builder=generator, **overrides
    )
    return functools.partial(
        GeneratedConfigEditor, editor_factory=config_editor_factory
    )


@deprecated("Use generate_device_configuration_editor instead")
def build_device_configuration_editor[
    C: DeviceConfiguration
](
    config_type: type[C],
    builder: EditorBuilder = _builder,
    **overrides: AttributeOverride,
) -> DeviceConfigEditorFactory[C]:
    return generate_device_configuration_editor(
        config_type, generator=builder, **overrides
    )


class DeviceServerNameEditor(ValueEditor[Optional[DeviceServerName]]):
    def __init__(self) -> None:
        self.line_edit = QLineEdit()
        self.line_edit.setPlaceholderText("None")

    def set_value(self, value: Optional[DeviceServerName]) -> None:
        if value:
            self.line_edit.setText(value)
        else:
            self.line_edit.setText("")

    def read_value(self) -> Optional[DeviceServerName]:
        text = self.line_edit.text()
        if text:
            return DeviceServerName(text)
        else:
            return None

    def set_editable(self, editable: bool) -> None:
        self.line_edit.setReadOnly(not editable)

    @property
    def widget(self) -> QLineEdit:
        return self.line_edit


class RectangularROIEditor(ValueEditor[RectangularROI]):
    def __init__(self) -> None:
        self._widget = RectangularROIWidget(100, 100)

    def set_value(self, value: RectangularROI) -> None:
        self._widget.set_roi(value)

    def read_value(self) -> RectangularROI:
        return self._widget.get_roi()

    def set_editable(self, editable: bool) -> None:
        self._widget.set_editable(editable)

    @property
    def widget(self) -> RectangularROIWidget:
        return self._widget


_builder.register_editor(Optional[DeviceServerName], DeviceServerNameEditor)
_builder.register_editor(RectangularROI, RectangularROIEditor)
_builder.register_editor(EvaluableOutput, OutputTransformEditor)
_builder.register_editor(Expression, ExpressionEditor)


def get_editor_builder() -> EditorBuilder:
    """Return a new editor builder with basic types registered.

    The editor builder returned knows how to handle:

        - Device server name
        - :class:`~caqtus.types.image.roi.RectangularROI`
        - :class:`~caqtus.types.expression.Expression`
    """

    return copy.deepcopy(_builder)



"""Defines types to be used for extending Condetrol with new lanes."""

from ._implementation import (
    CondetrolLaneExtension,
    LaneFactory,
    LaneModelFactory,
    LaneDelegateFactory,
)
from ._protocol import CondetrolLaneExtensionProtocol

__all__ = [
    "CondetrolLaneExtensionProtocol",
    "CondetrolLaneExtension",
    "LaneFactory",
    "LaneModelFactory",
    "LaneDelegateFactory",
]



from collections.abc import Callable

from caqtus.device import DeviceConfiguration, Device, DeviceController
from caqtus.device.configuration import DeviceServerName
from caqtus.device.remote import DeviceProxy, RPCConfiguration
from caqtus.shot_compilation import DeviceCompiler
from ._protocol import DeviceManagerExtensionProtocol


class DeviceManagerExtension(DeviceManagerExtensionProtocol):
    def __init__(self):
        self._compiler_types: dict[type[DeviceConfiguration], type[DeviceCompiler]] = {}
        self._device_types: dict[type[DeviceConfiguration], Callable[..., Device]] = {}
        self._controller_types: dict[
            type[DeviceConfiguration], type[DeviceController]
        ] = {}
        self._proxy_types: dict[type[DeviceConfiguration], type[DeviceProxy]] = {}
        self._device_server_configs: dict[DeviceServerName, RPCConfiguration] = {}

    def register_device_compiler(
        self,
        configuration_type: type[DeviceConfiguration],
        compiler_type: type[DeviceCompiler],
    ) -> None:
        self._compiler_types[configuration_type] = compiler_type

    def register_device[
        D: Device
    ](
        self,
        configuration_type: type[DeviceConfiguration[D]],
        device_type: Callable[..., D],
    ) -> None:
        self._device_types[configuration_type] = device_type

    def register_controller(
        self,
        configuration_type: type[DeviceConfiguration],
        controller_type: type[DeviceController],
    ) -> None:
        self._controller_types[configuration_type] = controller_type

    def register_proxy(
        self,
        configuration_type: type[DeviceConfiguration],
        proxy_type: type[DeviceProxy],
    ) -> None:
        self._proxy_types[configuration_type] = proxy_type

    def register_device_server_config(
        self,
        name: DeviceServerName,
        config: RPCConfiguration,
    ) -> None:
        self._device_server_configs[name] = config

    def get_device_compiler_type(
        self, device_configuration: DeviceConfiguration
    ) -> type[DeviceCompiler]:
        return self._compiler_types[type(device_configuration)]

    def get_device_type(
        self, device_configuration: DeviceConfiguration
    ) -> Callable[..., Device]:
        return self._device_types[type(device_configuration)]

    def get_device_controller_type(
        self, device_configuration: DeviceConfiguration
    ) -> type[DeviceController]:
        return self._controller_types[type(device_configuration)]

    def get_proxy_type(
        self, device_configuration: DeviceConfiguration
    ) -> type[DeviceProxy]:
        return self._proxy_types[type(device_configuration)]

    def get_device_server_config(self, server: DeviceServerName) -> RPCConfiguration:
        return self._device_server_configs[server]



from typing import TypeVar, NewType, Any, TypeGuard, TypeAlias

import numpy as np

from ._shape import Width, Height
from ..data import DataLabel, is_data_label

T = TypeVar("T", bound=np.generic)

Image: TypeAlias = np.ndarray[tuple[Width, Height], np.dtype[T]]

ImageLabel = NewType("ImageLabel", DataLabel)


def is_image(image: Any) -> TypeGuard[Image]:
    """Check if image has a valid image type."""

    return isinstance(image, np.ndarray) and image.ndim == 2


def is_image_label(label: Any) -> TypeGuard[ImageLabel]:
    """Check if label has a valid image label type."""

    return is_data_label(label)



from collections.abc import Iterable
from typing import Optional

from PySide6.QtWidgets import QDialog, QWidget

from .add_lane_dialog_ui import Ui_AddLaneDialog


class AddLaneDialog(QDialog, Ui_AddLaneDialog):
    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.setupUi(self)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.setWindowTitle("Add Lane...")

    def get_lane_name(self) -> str:
        return self.lane_name_line_edit.text()

    def get_lane_type(self) -> str:
        return self.lane_type_combobox.currentText()

    def set_lane_types(self, lane_types: Iterable[str]) -> None:
        self.lane_type_combobox.clear()
        self.lane_type_combobox.addItems(list(lane_types))



#!/usr/bin/python
# -*- coding: utf-8 -*-


class NodeMenuError(Exception):
    pass


class NodePropertyError(Exception):
    pass


class NodeWidgetError(Exception):
    pass


class NodeCreationError(Exception):
    pass


class NodeDeletionError(Exception):
    pass


class NodeRegistrationError(Exception):
    pass


class PortError(Exception):
    pass


class PortRegistrationError(Exception):
    pass



from collections.abc import Iterable
from typing import Self

import attrs
import numpy as np

from ._roi import ROI
from .._shape import Width, Height


@attrs.define
class ArbitraryROI(ROI):
    """Arbitrary region of interest inside an image.

    This ROI is defined by specifying the indices of the pixels in the original image
    that are part of the region of interest.

    While this ROI is the most generic one, it becomes very inefficient when the number
    of pixels in the region of interest is large, because it stores the coordinates of
    all pixels.
    Whenever possible, use a more specific ROI.

    Attributes:
        indices: The indices of the pixels in the original image that are part of the
        region of interest.
    """

    indices: tuple[tuple[int, int], ...] = attrs.field(
        converter=tuple,
        on_setattr=attrs.setters.pipe(attrs.setters.convert, attrs.setters.validate),
    )

    def get_indices(self) -> tuple[Iterable[int], Iterable[int]]:
        indices = np.array(self.indices).T
        return indices[0], indices[1]

    @classmethod
    def from_mask(cls, mask: np.ndarray) -> Self:
        """Create a region of interest from a mask

        Args:
            mask: A boolean array with the same shape as the original image.
            True values indicate that the pixel is part of the region of interest.
        """

        shape = mask.shape
        if len(shape) != 2:
            raise ValueError("mask must be 2D")
        shape = Width(shape[0]), Height(shape[1])
        indices = np.argwhere(mask).tolist()
        return cls(
            original_image_size=shape, indices=tuple(tuple(index) for index in indices)
        )

    @indices.validator  # type: ignore
    def validate_indices(self, _, indices):
        for index in indices:
            if not isinstance(index, tuple):
                raise ValueError("indices must be a list of tuples")
            if len(index) != 2:
                raise ValueError("indices must be a list of pairs")
            inside = (
                0 <= index[0] < self.original_image_size[0]
                and 0 <= index[1] < self.original_image_size[1]
            )
            if not inside:
                raise ValueError("indices must be inside the original image")
        return indices



import decimal
from typing import NewType

from caqtus.shot_compilation.timing import start_tick, stop_tick, number_ticks, Time, ps

__all__ = [
    "TimeStep",
    "start_time_step",
    "stop_time_step",
    "number_time_steps",
    "number_time_steps_between",
    "to_time_step",
    "ns",
]

TimeStep = NewType("TimeStep", decimal.Decimal)
"""A type alias that represents the duration of a time step in nanoseconds.

A time step is represented as a decimal number to avoid floating point errors.
"""

ns = Time(decimal.Decimal("1e-9"))


def to_time_step(value: decimal.Decimal | float | str) -> TimeStep:
    """Converts a value to a TimeStep object.

    Args:
        value: The value to convert to a TimeStep object, in nanoseconds.

    Returns:
        A TimeStep object representing the value in nanoseconds, rounded to the
        picosecond.
    """

    return TimeStep(decimal.Decimal(value).quantize(ps / ns))


def start_time_step(start_time: Time, time_step: TimeStep) -> int:
    """Returns the time of the step starting at start_time."""

    return start_tick(start_time, Time(time_step * ns))


def stop_time_step(stop_time: Time, time_step: TimeStep) -> int:
    """Returns the time of the step ending at stop_time."""

    return stop_tick(stop_time, Time(time_step * ns))


def number_time_steps(duration: Time, time_step: TimeStep) -> int:
    """Returns the number of ticks covering the given duration.

    Args:
        duration: The duration in seconds.
        time_step: The time step in seconds.
    """

    return number_ticks(Time(decimal.Decimal(0)), duration, Time(time_step * ns))


def number_time_steps_between(
    start_time: Time, stop_time: Time, time_step: TimeStep
) -> int:
    """Returns the number of ticks covering the given duration.

    Args:
        start_time: The start time in seconds.
        stop_time: The stop time in seconds.
        time_step: The time step in seconds.
    """

    return number_ticks(start_time, stop_time, Time(time_step * ns))



from __future__ import annotations

from typing import Optional, Any

import attrs
import numpy as np
import polars

from caqtus.session import Shot
from caqtus.types.image import ImageLabel, is_image
from caqtus.types.image.roi import ROI, RectangularROI
from ._combinable_importers import CombinableLoader
from ...types.data import DataLabel
from ...types.recoverable_exceptions import InvalidTypeError


@attrs.define
class LoadImageCount(CombinableLoader):
    """Loads the count values of an image.

    This loader computes the sum of the pixel values of an image.

    Args:
        camera_name: The name of the camera.
        picture_name: The name of the picture.
        background_name: The name of the background picture to subtract before computing
            the sum. If None, no background is subtracted.
        roi: The region of interest to reduce the sum to.
            If None, the whole picture is used.
        column_name: The name of the column to store the computed sum in.
            If not provided, the column is named '{picture_name} count'.
    """

    camera_name: str
    picture_name: ImageLabel
    background_name: Optional[ImageLabel] = None
    roi: Optional[ROI] = None
    column_name: Optional[str] = None

    _indices: Any = attrs.field(init=False, default=None)
    _image_data_label: str = attrs.field(init=False, default=None)
    _background_data_label: Optional[str] = attrs.field(init=False, default=None)

    def __attrs_post_init__(self):
        self._column_name = self.column_name or f"{self.picture_name} count"
        self._image_data_label = rf"{self.camera_name}\{self.picture_name}"
        if self.background_name is not None:
            self._background_data_label = rf"{self.camera_name}\{self.background_name}"
        match self.roi:
            case None:
                self._indices = (slice(None), slice(None))
            case RectangularROI() as rectangular_roi:
                self._indices = (
                    slice(rectangular_roi.x, rectangular_roi.x + rectangular_roi.width),
                    slice(
                        rectangular_roi.y, rectangular_roi.y + rectangular_roi.height
                    ),
                )
            case _:
                self._indices = self.roi.get_indices()

    def load(self, shot: Shot) -> polars.DataFrame:
        image = shot.get_data_by_label(DataLabel(self._image_data_label))
        if not is_image(image):
            raise InvalidTypeError(
                f"Expected data '{self._image_data_label}' to be an image, got {image} "
                f"instead"
            )

        image = image[self._indices]

        picture_count = np.sum(image)
        if self._background_data_label is None:
            background_count = 0
        else:
            background = shot.get_data_by_label(DataLabel(self._background_data_label))
            if not is_image(background):
                raise InvalidTypeError(
                    f"Expected data '{self._background_data_label}' to be an image, "
                    f"got {background} instead"
                )
            background_count = np.sum(background)

        count = picture_count - background_count
        return polars.DataFrame(polars.Series(self._column_name, [count]))



import itertools
from collections.abc import Mapping
from typing import assert_never

import attrs

from caqtus.types.parameter import Parameter
from caqtus.types.units import Unit, Quantity
from caqtus.types.variable_name import DottedVariableName

type ConstantSchema = Mapping[DottedVariableName, Parameter]
type VariableSchema = Mapping[DottedVariableName, ParameterType]
type ParameterType = (Boolean | Integer | Float | QuantityType)


class ParameterSchema(Mapping[DottedVariableName | str, ParameterType]):
    """Contains the type of each parameter in a sequence."""

    def __init__(
        self,
        *,
        _constant_schema: ConstantSchema,
        _variable_schema: VariableSchema,
    ) -> None:
        if set(_constant_schema) & set(_variable_schema):
            raise ValueError(
                "The constant and variable schemas must not have any parameters in "
                "common."
            )
        self._constant_schema = _constant_schema
        self._variable_schema = _variable_schema

    def __len__(self):
        return len(self._constant_schema) + len(self._variable_schema)

    def __iter__(self):
        return itertools.chain(self._constant_schema, self._variable_schema)

    def __getitem__(self, key: DottedVariableName | str) -> ParameterType:
        if isinstance(key, str):
            key = DottedVariableName(key)
        if key in self._constant_schema:
            return self.type_from_value(self._constant_schema[key])
        elif key in self._variable_schema:
            return self._variable_schema[key]
        else:
            raise KeyError(key)

    @property
    def constant_schema(self) -> ConstantSchema:
        """Values of the parameters that are constant during the sequence."""

        return self._constant_schema

    @property
    def variable_schema(self) -> VariableSchema:
        """Types of the parameters that can change during the sequence."""

        return self._variable_schema

    def __repr__(self) -> str:
        return (
            f"ParameterSchema("
            f"_constant_schema={self._constant_schema}, "
            f"_variable_schema={self._variable_schema})"
        )

    def __str__(self) -> str:
        constants = (
            f'"{key}": {value}' for key, value in self._constant_schema.items()
        )
        variables = (
            f'"{key}": {value}' for key, value in self._variable_schema.items()
        )
        joined = itertools.chain(constants, variables)
        return "{" + ", ".join(joined) + "}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ParameterSchema):
            return NotImplemented
        return (
            self._constant_schema == other._constant_schema
            and self._variable_schema == other._variable_schema
        )

    @classmethod
    def type_from_value(cls, value: Parameter) -> ParameterType:
        if isinstance(value, bool):
            return Boolean()
        elif isinstance(value, int):
            return Integer()
        elif isinstance(value, float):
            return Float()
        elif isinstance(value, Quantity):
            return QuantityType(units=value.units)
        else:
            assert_never(value)


@attrs.frozen
class QuantityType:
    units: Unit


@attrs.frozen
class Float:
    @property
    def units(self) -> None:
        return None


@attrs.frozen
class Boolean:
    @property
    def units(self) -> None:
        return None


@attrs.frozen
class Integer:
    @property
    def units(self) -> None:
        return None



from typing import Protocol, Optional, TypeVar

from caqtus.types.timelane import TimeLanes, TimeLane
from caqtus.utils import serialization
from .._delegate import TimeLaneDelegate
from .._time_lane_model import TimeLaneModel

L = TypeVar("L", bound=TimeLane)


class CondetrolLaneExtensionProtocol(Protocol):
    """Defines the operations necessary to extend Condetrol with new lanes."""

    def available_new_lanes(self) -> set[str]:
        """Return the new lanes that can be created.

        This method is called when the user clicks on the "Add lane" button and needs
        to choose the type of lane to create.
        """

        ...

    def create_new_lane(self, lane_label: str, steps: int) -> TimeLane:
        """Create a new lane.

        This method is called when the user wants to create a new lane.
        The label of the lane to create and the number of steps are passed as arguments.
        """

        ...

    def get_lane_model(self, lane: L, name: str) -> TimeLaneModel[L]:
        """Return the model for the given lane.

        This method is called when a lane needs to be displayed.
        The returned model will be used to provide the data from the lane to the view.
        """

        ...

    def get_lane_delegate(
        self,
        lane: TimeLane,
        lane_name: str,
    ) -> Optional[TimeLaneDelegate]:
        """Return a delegate for the given lane.

        This method is called when a lane needs to be displayed.
        The returned delegate will be used to paint the lane cells in the view and to
        provide editing capabilities.
        """

        ...

    def unstructure_time_lanes(self, time_lanes: TimeLanes) -> serialization.JsonDict:
        """Convert the time lanes to a serializable format."""

        ...

    def structure_time_lanes(self, content: serialization.JsonDict) -> TimeLanes:
        """Convert the serializable format to time lanes."""

        ...



from __future__ import annotations

from collections.abc import Iterable
from typing import Self, Any

import attrs

from caqtus.utils import serialization


@attrs.define(repr=False, eq=False, init=False)
class DottedVariableName:
    """Represents the name of a parameter.

    Instances of this class represents the name of a parameter.
    They must be valid python identifiers, possibly separated by dots to represent
    a hierarchy of names.

    For example, the names "a", "foo.bar", and "a.b.c" are all valid names.
    """

    _dotted_name: str
    _individual_names: tuple[VariableName, ...] = attrs.field(init=False, repr=False)

    def __init__(self, dotted_name: str):
        names = tuple(dotted_name.split("."))
        self._individual_names = tuple(VariableName(name) for name in names)
        self._dotted_name = str(dotted_name)

    @property
    def dotted_name(self) -> str:
        """The string representation of the name."""

        return self._dotted_name

    @property
    def individual_names(self) -> tuple[VariableName, ...]:
        """The individual separated by dots that make up the identifier."""

        return self._individual_names

    @classmethod
    def from_individual_names(cls, names: Iterable[VariableName]) -> Self:
        """Create a new instance from an iterable of individual names."""

        return cls(".".join(str(name) for name in names))

    def __str__(self) -> str:
        return self._dotted_name

    def __repr__(self):
        return f"{type(self).__name__}('{self._dotted_name}')"

    def __hash__(self):
        return hash(self._dotted_name)

    def __eq__(self, other):
        if isinstance(other, DottedVariableName):
            return self._dotted_name == other._dotted_name
        elif isinstance(other, str):
            return self._dotted_name == other
        else:
            return NotImplemented


def dotted_variable_name_converter(name: Any) -> DottedVariableName:
    if isinstance(name, DottedVariableName):
        return name
    elif isinstance(name, str):
        return DottedVariableName(name)
    else:
        raise InvalidVariableNameError(f"Invalid variable name: {name}")


@attrs.define(eq=False, repr=False, init=False)
class VariableName(DottedVariableName):
    """Represents a single variable name."""

    def __init__(self, name: str):
        if not name.isidentifier():
            raise InvalidVariableNameError(f"Invalid variable name: {name}")
        self._individual_names = (self,)
        self._dotted_name = str(name)

    def __hash__(self):
        return super().__hash__()

    def __repr__(self):
        return super().__repr__()

    def __str__(self):
        return super().__str__()

    def __eq__(self, other):
        if isinstance(other, VariableName):
            return self._dotted_name == other._dotted_name
        elif isinstance(other, str):
            return self._dotted_name == other
        else:
            return NotImplemented


def unstructure_hook(dotted_variable_name: DottedVariableName) -> str:
    return dotted_variable_name.dotted_name


serialization.register_unstructure_hook(DottedVariableName, unstructure_hook)


def structure_hook(data: str, cls: type[DottedVariableName]) -> DottedVariableName:
    return DottedVariableName(data)


serialization.register_structure_hook(DottedVariableName, structure_hook)


class InvalidVariableNameError(ValueError):
    pass



import functools
from collections.abc import Set
from typing import Optional

import yaml
from PySide6 import QtWidgets
from PySide6.QtCore import Qt, QPersistentModelIndex, QModelIndex
from PySide6.QtGui import (
    QKeySequence,
    QShortcut,
    QFont,
    QGuiApplication,
    QUndoStack,
)
from PySide6.QtWidgets import QWidget, QTreeView, QAbstractItemView, QMenu

from caqtus.gui.qtutil._temporary_widget import temporary_widget
from caqtus.types.expression import Expression
from caqtus.types.iteration import (
    StepsConfiguration,
    VariableDeclaration,
    ExecuteShot,
    LinspaceLoop,
    ArangeLoop,
)
from caqtus.types.variable_name import DottedVariableName
from caqtus.utils import serialization
from .delegate import StepDelegate
from .steps_model import StepsModel
from ..sequence_iteration_editor import SequenceIterationEditor
from ..._icons import get_icon


def create_shot_step() -> ExecuteShot:
    return ExecuteShot()


def create_variable_declaration():
    return VariableDeclaration(
        variable=DottedVariableName("new_variable"), value=Expression("...")
    )


def create_linspace_loop():
    return LinspaceLoop(
        variable=DottedVariableName("new_variable"),
        start=Expression("..."),
        stop=Expression("..."),
        num=10,
        sub_steps=[],
    )


def create_arange_loop():
    return ArangeLoop(
        variable=DottedVariableName("new_variable"),
        start=Expression("..."),
        stop=Expression("..."),
        step=Expression("..."),
        sub_steps=[],
    )


class StepsIterationEditor(QTreeView, SequenceIterationEditor[StepsConfiguration]):
    """Editor for the steps of a sequence.

    .. figure:: screenshot_StepsIterationEditor.png
       :align: center

       Example of a StepsIterationEditor.

    Args:
        parent: The parent widget of the editor.

    Signals:
        iteration_edited: Emitted when the iteration displayed by the editor is changed
            by the user.
            This signal is not emitted when the iteration is set programmatically.
    """

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self._model = StepsModel(StepsConfiguration([]))
        self.setModel(self._model)
        self.expandAll()
        self.header().hide()
        self._delegate = StepDelegate(self)
        self.setItemDelegate(self._delegate)

        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setDropIndicatorShown(True)
        self.setDragDropMode(QAbstractItemView.DragDropMode.DragDrop)
        self.setDefaultDropAction(Qt.DropAction.MoveAction)

        self.delete_shortcut = QShortcut(QKeySequence("Delete"), self)
        self.delete_shortcut.activated.connect(self.delete_selected)

        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.show_context_menu)

        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

        self.select_all_shortcut = QShortcut(QKeySequence("Ctrl+A"), self)
        self.select_all_shortcut.activated.connect(self.selectAll)
        self.copy_shortcut = QShortcut(QKeySequence("Ctrl+C"), self)
        self.copy_shortcut.activated.connect(self.copy_selection)
        self.cut_shortcut = QShortcut(QKeySequence("Ctrl+X"), self)
        self.cut_shortcut.activated.connect(self.cut_selected)
        self.paste_shortcut = QShortcut(QKeySequence("Ctrl+V"), self)
        self.paste_shortcut.activated.connect(self.paste)

        font = QFont("JetBrains Mono")
        font.setPixelSize(13)
        self.setFont(font)

        self.toolbar = QtWidgets.QToolBar()
        self.add_button = QtWidgets.QToolButton(self)
        self.add_menu = QMenu(self)
        self.add_shot_action = self.add_menu.addAction("shot")
        self.add_variable_action = self.add_menu.addAction("variable")
        self.add_linspace_action = self.add_menu.addAction("linspace loop")
        self.add_arange_action = self.add_menu.addAction("arange loop")
        self._setup_add_button()
        self.toolbar.addWidget(self.add_button)

        self.toolbar.addSeparator()
        self.copy_to_clipboard_action = self.toolbar.addAction(
            get_icon("copy", self.palette().buttonText().color()), "Copy to clipboard"
        )
        self.copy_to_clipboard_action.triggered.connect(self.copy_to_clipboard)
        self.paste_from_clipboard_action = self.toolbar.addAction(
            get_icon("paste", self.palette().buttonText().color()),
            "Paste from clipboard",
        )
        self.paste_from_clipboard_action.triggered.connect(self.paste_from_clipboard)

    def has_uncommitted_edits(self) -> bool:
        return not self.undo_stack.isClean()

    def commit_edits(self) -> None:
        self.undo_stack.setClean()

    @property
    def undo_stack(self) -> QUndoStack:
        """The undo stack used by the editor.

        This undo stack is used to undo and redo the changes made by the user in the
        editor.
        """

        return self._model.undo_stack

    def _setup_add_button(self):
        self.add_button.setToolTip("Add step")
        self.add_button.setPopupMode(
            QtWidgets.QToolButton.ToolButtonPopupMode.InstantPopup
        )
        self.add_button.setIcon(get_icon("plus", self.palette().buttonText().color()))
        self.add_button.setMenu(self.add_menu)
        self.add_shot_action.triggered.connect(
            functools.partial(
                self._model.insert_steps,
                [create_shot_step()],
                self._model.rowCount(),
                QModelIndex(),
            )
        )
        self.add_variable_action.triggered.connect(
            functools.partial(
                self._model.insert_steps,
                [create_variable_declaration()],
                0,
                QModelIndex(),
            )
        )
        self.add_linspace_action.triggered.connect(
            functools.partial(
                self._model.insert_steps, [create_linspace_loop()], 0, QModelIndex()
            )
        )
        self.add_arange_action.triggered.connect(
            functools.partial(
                self._model.insert_steps, [create_arange_loop()], 0, QModelIndex()
            )
        )

    def get_iteration(self) -> StepsConfiguration:
        """Returns the iteration currently displayed by the editor.

        This method returns a copy of the iteration, so the caller can modify it
        without affecting the editor.
        """

        return self._model.get_steps()

    def set_iteration(self, iteration: StepsConfiguration):
        """Updates the iteration displayed by the editor.

        This method does not emit the signal iteration_edited.
        """

        self._model.set_steps(iteration)
        self.expandAll()

    def set_available_parameter_names(self, parameter_names: Set[DottedVariableName]):
        """Sets the names that can be used in the iteration.

        The names are used to populate the auto-completion of variable names when the
        user try to edit the name of a variable or a loop.
        """

        self._delegate.set_available_names(parameter_names)

    def set_read_only(self, read_only: bool) -> None:
        """Sets the editor in read-only mode.

        When the editor is in read-only mode, the user cannot edit the steps.

        Even if the editor is in read-only mode, the iteration can still be set
        programmatically with :meth:`set_iteration`.
        """

        self._model.set_read_only(read_only)
        self.paste_from_clipboard_action.setEnabled(not read_only)
        self.add_button.setEnabled(not read_only)

    def is_read_only(self) -> bool:
        """Returns whether the editor is in read-only mode."""

        return self._model.is_read_only()

    def delete_selected(self) -> bool:
        if self.is_read_only():
            return False
        self._model.remove_indices(self.selectedIndexes())
        return True

    def copy_selection(self) -> None:
        selected_indexes = self.selectedIndexes()
        data = self._model.mimeData(selected_indexes)
        clipboard = QGuiApplication.clipboard()
        clipboard.setMimeData(data)

    def cut_selected(self) -> bool:
        if self.is_read_only():
            return False

        self.copy_selection()
        self._model.remove_indices(self.selectedIndexes())
        return True

    def paste(self) -> bool:
        if self.is_read_only():
            return False

        clipboard = QGuiApplication.clipboard()
        data = clipboard.mimeData()

        if self.selectedIndexes():
            first = QPersistentModelIndex(self.selectedIndexes()[0])
            parent = first.parent()
            row = first.row()
            if self._model.dropMimeData(data, Qt.DropAction.MoveAction, row, 0, parent):
                self._model.removeRow(first.row(), first.parent())
        else:
            parent = QModelIndex()
            row = self._model.rowCount()
            self._model.dropMimeData(data, Qt.DropAction.MoveAction, row, 0, parent)
        return True

    def show_context_menu(self, position):
        index = self.indexAt(position)
        if not index.isValid():
            return
        with temporary_widget(QMenu(self)) as menu:
            add_menu = menu.addMenu("Insert above...")

            create_variable_action = add_menu.addAction("variable")
            new_variable = create_variable_declaration()
            create_variable_action.triggered.connect(
                functools.partial(
                    self._model.insert_steps,
                    [new_variable],
                    index.row(),
                    index.parent(),
                )
            )
            create_shot_action = add_menu.addAction("Shot")
            new_shot = ExecuteShot()
            create_shot_action.triggered.connect(
                functools.partial(
                    self._model.insert_steps, [new_shot], index.row(), index.parent()
                )
            )
            create_linspace_action = add_menu.addAction("Linspace loop")
            new_linspace = create_linspace_loop()
            create_linspace_action.triggered.connect(
                functools.partial(
                    self._model.insert_steps,
                    [new_linspace],
                    index.row(),
                    index.parent(),
                )
            )
            create_arange_action = add_menu.addAction("Arange loop")
            new_arange = create_arange_loop()
            create_arange_action.triggered.connect(
                functools.partial(
                    self._model.insert_steps, [new_arange], index.row(), index.parent()
                )
            )
            menu.exec(self.mapToGlobal(position))

    def copy_to_clipboard(self):
        steps = self.get_iteration()

        unstructured = serialization.converters["json"].unstructure(
            steps, StepsConfiguration
        )

        text = yaml.safe_dump(unstructured)
        clipboard = QGuiApplication.clipboard()
        clipboard.setText(text)

    def paste_from_clipboard(self) -> bool:
        """Attempts to paste the content of the clipboard.

        If the editor is in read-only mode, this method does nothing.
        """

        if self._model.is_read_only():
            return False

        clipboard = QGuiApplication.clipboard()
        text = clipboard.text()
        try:
            data = yaml.safe_load(text)
        except yaml.YAMLError as e:
            QtWidgets.QMessageBox.warning(  # type: ignore[reportCallIssue]
                self,
                "Invalid YAML content",
                f"Could not parse the clipboard content as YAML:\n {e}",
            )
            return False

        # TODO: raise recoverable exception if data is not valid
        config = serialization.structure(data, StepsConfiguration)
        self._model.replace_steps(config.steps, "paste from clipboard")

        return True



from collections.abc import Mapping

from caqtus.types.units import Quantity, Unit
from ._scalar import Scalar

CONSTANTS: Mapping[str, Scalar] = {
    "pi": Quantity(3.141592653589793, Unit("rad")),
    "e": 2.718281828459045,
    "Enabled": True,
    "Disabled": False,
}



import warnings

from caqtus.shot_compilation.timed_instructions import (
    TimedInstruction,
    Concatenated,
    Repeated,
    Pattern,
    convert_to_change_arrays,
    with_name,
    stack_instructions,
    merge_instructions,
    concatenate,
    create_ramp,
    Ramp,
    plot_instruction,
    to_graph,
)

warnings.warn(
    "caqtus.device.sequencer.instructions is deprecated, use "
    "caqtus.shot_compilation.timed_instructions instead.",
    DeprecationWarning,
    stacklevel=2,
)

SequencerInstruction = TimedInstruction

__all__ = [
    "SequencerInstruction",
    "TimedInstruction",
    "Concatenated",
    "Repeated",
    "Pattern",
    "convert_to_change_arrays",
    "with_name",
    "stack_instructions",
    "merge_instructions",
    "concatenate",
    "create_ramp",
    "Ramp",
    "plot_instruction",
    "to_graph",
]



import attrs

from ._protocol import CondetrolExtensionProtocol
from ..device_configuration_editors._extension import CondetrolDeviceExtension
from ..timelanes_editor.extension import CondetrolLaneExtension


@attrs.define
class CondetrolExtension(CondetrolExtensionProtocol):
    _lane_extension: CondetrolLaneExtension = attrs.field(
        factory=CondetrolLaneExtension
    )
    _device_extension: CondetrolDeviceExtension = attrs.field(
        factory=CondetrolDeviceExtension
    )

    @property
    def lane_extension(self) -> CondetrolLaneExtension:
        return self._lane_extension

    @property
    def device_extension(self) -> CondetrolDeviceExtension:
        return self._device_extension



from PySide6.QtWidgets import QLineEdit

from caqtus.types.expression import Expression
from ._value_editor import ValueEditor


class ExpressionEditor(ValueEditor[Expression]):
    def __init__(self) -> None:
        self._line_edit = QLineEdit()
        self._line_edit.setPlaceholderText("Variable or math expression")

    def set_value(self, value: Expression) -> None:
        self._line_edit.setText(str(value))

    def read_value(self) -> Expression:
        return Expression(self._line_edit.text())

    def set_editable(self, editable: bool) -> None:
        self._line_edit.setReadOnly(not editable)

    @property
    def widget(self) -> QLineEdit:
        return self._line_edit



"""Defines the result type and its variants: success and failure.

The result type is a union type of :class:`Success` and :class:`Failure` respectively
containing a successful value or an error code.

It is mostly meant to be used as a return type for functions that can fail, but where
we want to be sure to handle all cases in the calling code and not raise unhandled
exceptions.

With a type checker, we can ensure that all possible success and failure cases are
dealt with.

Example:
    .. code-block:: python

        from typing import assert_never

        from caqtus.utils.result import Success, Failure, is_success, is_failure_type

        def read_file(path) -> Success[str] | Failure[FileNotFoundError]:
            try:
                with open(path) as file:
                    return Success(file.read())
            except FileNotFoundError as error:
                return Failure(error)

        result = read_file("file.txt")
        if is_failure_type(result, FileNotFoundError):
            print("File not found")
        elif is_success(result):
            print(result.content())
        else:
            assert_never(result)
"""

from ._result import (
    Failure,
    Success,
    is_failure,
    is_failure_type,
    is_success,
    unwrap,
)

__all__ = [
    "Failure",
    "Success",
    "is_failure",
    "is_failure_type",
    "is_success",
    "unwrap",
]



from typing import Optional

import attrs

from caqtus.gui.condetrol.timelanes_editor.extension import (
    LaneFactory,
    LaneModelFactory,
    LaneDelegateFactory,
)
from caqtus.types.timelane import TimeLane, LaneLoader, LaneDumper


def no_lane_delegate_factory(lane: TimeLane, lane_name: str) -> None:
    return None


@attrs.frozen
class TimeLaneExtension[L: TimeLane]:
    """Define how to implement a time lane plugin.

    Attributes:
        label: An identifier for this type of lane to be displayed to the user.
        lane_type: The type of lane to be created.
        dumper: A function to serialize the lane to JSON.
            When a lane with the corresponding type needs to be saved, this function
            will be called and the result will be stored.
            The returned value must be a dictionary that can be serialized to JSON.
            The dictionary will be added a "type" key to identify the lane type.
        loader: A function to deserialize the lane from JSON.
            When JSON data with the corresponding "type" key is loaded, this function
            will be called to create a lane.
        lane_factory: A factory function to create a new lane when the user wants to
            create a lane with this label.
            The factory will be called with the number of steps the lane must have.
        lane_delegate_factory: A factory function to create a delegate for the lane.
            The factory will be called when the lane is displayed in the editor.
            The delegate returned by the factory will be used for custom painting and
            editing of the lane.
            The default lane delegate factory returns None.
        lane_model_factory: A factory function to create a model for the lane.
            The model will be used to provide the data from the lane to the view.
    """

    label: str = attrs.field(converter=str)
    lane_type: type[L] = attrs.field()
    dumper: LaneDumper[L] = attrs.field()
    loader: LaneLoader[L] = attrs.field()
    lane_factory: LaneFactory[L] = attrs.field()
    lane_model_factory: LaneModelFactory[L] = attrs.field()
    lane_delegate_factory: LaneDelegateFactory[L] = attrs.field(
        default=no_lane_delegate_factory
    )
    type_tag: Optional[str] = attrs.field(default=None)

    @lane_type.validator  # type: ignore[reportAttributeAccessIssue]
    def _validate_lane_type(self, attribute, value):
        if not issubclass(value, TimeLane):
            raise ValueError(f"{value} is not a subclass of TimeLane")



from collections.abc import Iterable
from typing import Optional

from PySide6.QtCore import (
    QObject,
    Qt,
    Signal,
    QModelIndex,
    QMimeData,
    QEvent,
    QPersistentModelIndex,
)
from PySide6.QtGui import (
    QStandardItemModel,
    QStandardItem,
    QPalette,
    QFocusEvent,
    QFont,
    QUndoStack,
)
from PySide6.QtWidgets import (
    QWidget,
    QColumnView,
    QSizePolicy,
    QApplication,
    QToolBar,
    QToolButton,
    QMenu,
    QVBoxLayout,
    QAbstractItemView,
    QHBoxLayout,
    QLabel,
    QStyleOptionViewItem,
)

from caqtus.gui._common.exception_tree import ExceptionDialog
from caqtus.gui.qtutil import block_signals, temporary_widget
from caqtus.session import TracebackSummary
from caqtus.types.expression import Expression
from caqtus.types.parameter import ParameterNamespace
from caqtus.types.variable_name import DottedVariableName
from caqtus.utils import serialization
from .._icons import get_icon
from .._logger import logger
from .._qt_util import AutoResizeLineEdit
from ...qtutil import HTMLItemDelegate

type AnyModelIndex = QModelIndex | QPersistentModelIndex

logger = logger.getChild("parameters_editor")

PARAMETER_NAME_ROLE = Qt.ItemDataRole.UserRole + 1
PARAMETER_VALUE_ROLE = Qt.ItemDataRole.UserRole + 2

DEFAULT_INDEX = QModelIndex()


class ParameterNamespaceEditor(QWidget):
    """A widget that allows to edit a ParameterNamespace.

    This widget presents a column view with the parameters and namespaces in each
    column.
    It also has a toolbar with buttons to add, remove, copy and paste parameters.
    """

    # The argument is a ParameterNamespace, but this is not a valid type for the
    # Signal.
    parameters_edited = Signal(object)

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.view = ParameterNamespaceView(self)

        self.tool_bar = QToolBar(self)
        self.undo_stack = QUndoStack(self)

        self._model = ParameterNamespaceModel(self)
        self.view.setModel(self._model)

        self.add_button = QToolButton(self)
        self.add_menu = QMenu(self)
        self.add_parameter_action = self.add_menu.addAction("Add parameter")
        self.add_namespace_action = self.add_menu.addAction("Add namespace")

        self.delete_button = QToolButton(self)
        self.copy_to_clipboard_button = QToolButton(self)
        self.paste_from_clipboard_button = QToolButton(self)

        self.setup_ui()
        self.setup_connections()
        self.set_read_only(False)

        font = QFont("JetBrains Mono")
        self.setFont(font)

    def setup_ui(self):
        layout = QVBoxLayout(self)
        layout.addWidget(self.view)
        layout.addWidget(self.tool_bar)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        self.copy_to_clipboard_button.clicked.connect(
            self.on_copy_to_clipboard_button_clicked
        )
        self.paste_from_clipboard_button.clicked.connect(
            self.on_paste_from_clipboard_button_clicked
        )
        self.add_button.setMenu(self.add_menu)
        self.add_button.setToolTip("Add")
        self.add_button.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)
        self.delete_button.setToolTip("Remove")
        self.copy_to_clipboard_button.setToolTip("Copy to clipboard")
        self.paste_from_clipboard_button.setToolTip("Paste from clipboard")
        self.tool_bar.addWidget(self.add_button)
        self.tool_bar.addWidget(self.delete_button)
        self.tool_bar.addSeparator()
        self.tool_bar.addWidget(self.copy_to_clipboard_button)
        self.tool_bar.addWidget(self.paste_from_clipboard_button)

        self.set_parameters(ParameterNamespace.empty())

    def setup_connections(self) -> None:
        def emit_edited_signal(*_):
            self.parameters_edited.emit(self.get_parameters())

        self._model.dataChanged.connect(emit_edited_signal)
        self._model.modelReset.connect(emit_edited_signal)
        self._model.rowsInserted.connect(emit_edited_signal)
        self._model.rowsRemoved.connect(emit_edited_signal)
        self._model.rowsMoved.connect(emit_edited_signal)
        self.delete_button.clicked.connect(self.on_delete_button_clicked)
        self.add_parameter_action.triggered.connect(
            lambda: self._model.add_parameter(
                DottedVariableName("new_parameter"), Expression("...")
            )
        )
        self.add_namespace_action.triggered.connect(
            lambda: self._model.add_namespace(DottedVariableName("new_namespace"))
        )

    def set_read_only(self, read_only: bool) -> None:
        if read_only:
            self.on_set_to_read_only()
        else:
            self.on_set_to_editable()

    def on_set_to_read_only(self) -> None:
        self.add_button.setEnabled(False)
        self.delete_button.setEnabled(False)
        self.paste_from_clipboard_button.setEnabled(False)
        self._model.set_read_only(True)

    def on_set_to_editable(self) -> None:
        self.add_button.setEnabled(True)
        self.delete_button.setEnabled(True)
        self.paste_from_clipboard_button.setEnabled(True)
        self._model.set_read_only(False)

    def on_delete_button_clicked(self) -> None:
        """Remove the selected item."""

        index = self.view.currentIndex()
        self._model.removeRow(index.row(), index.parent())

    def set_parameters(self, parameters: ParameterNamespace) -> None:
        """Set the parameters to be displayed in the table.

        This method ignore the read-only flag and always set the parameters displayed.
        It does not emit the parameters_edited signal.
        """

        with block_signals(self):
            self._set_parameters(parameters)

    def _set_parameters(self, parameters: ParameterNamespace) -> None:
        # The palette is not set yet in the __init__, so we need to update the icons
        # here, now that it is set to have the right color.
        color = self.palette().buttonText().color()
        self.add_button.setIcon(get_icon("plus", color))
        self.delete_button.setIcon(get_icon("minus", color))
        self.copy_to_clipboard_button.setIcon(get_icon("copy", color))
        self.paste_from_clipboard_button.setIcon(get_icon("paste", color))
        self._model.set_parameters(parameters)

    def get_parameters(self) -> ParameterNamespace:
        """Return the parameters displayed in the table."""

        return self._model.get_parameters()

    def on_copy_to_clipboard_button_clicked(self) -> None:
        """Copy all the displayed parameters to the clipboard."""

        parameters = self.get_parameters()
        serialized = serialization.to_json(parameters, ParameterNamespace)
        clipboard = QApplication.clipboard()
        clipboard.setText(serialized)

    def on_paste_from_clipboard_button_clicked(self) -> None:
        """Paste the parameters from the clipboard and display them in the table."""

        clipboard = QApplication.clipboard()
        serialized = clipboard.text()
        try:
            parameters = serialization.from_json(serialized, ParameterNamespace)
        except Exception as e:
            with temporary_widget(ExceptionDialog(self)) as dialog:
                dialog.set_exception(TracebackSummary.from_exception(e))
                dialog.set_message("The clipboard does not contain valid parameters.")
                dialog.exec()
        else:
            self._set_parameters(parameters)


class ParameterNamespaceView(QColumnView):
    """A custom QColumnView used to display the values in a parameter namespace."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Fix to hide the preview widget column
        # see: https://bugreports.qt.io/browse/QTBUG-1826
        self.w = QWidget()
        self.w.setMaximumSize(0, 0)
        self.w.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.setPreviewWidget(self.w)
        self.updatePreviewWidget.connect(self._on_update_preview_widget)

        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setDropIndicatorShown(True)
        self.setDragDropMode(QAbstractItemView.DragDropMode.DragDrop)
        self.setDefaultDropAction(Qt.DropAction.MoveAction)
        self.setResizeGripsVisible(True)

        self.delegate = ParameterEditorDelegate(self)
        self.setItemDelegate(self.delegate)
        self.setEditTriggers(
            QAbstractItemView.EditTrigger.NoEditTriggers
            | QAbstractItemView.EditTrigger.DoubleClicked
            | QAbstractItemView.EditTrigger.EditKeyPressed
        )
        self.setSelectionMode(QColumnView.SelectionMode.SingleSelection)
        self.setSelectionBehavior(QColumnView.SelectionBehavior.SelectItems)

    def _on_update_preview_widget(self, index):
        parent = self.w.parentWidget()
        assert parent is not None
        grand_parent = parent.parentWidget()
        assert grand_parent is not None
        grand_parent.setMinimumWidth(0)
        grand_parent.setMaximumWidth(0)


class ParameterNamespaceModel(QStandardItemModel):
    # ruff: noqa: N802
    def __init__(self, parent: Optional[QObject] = None):
        super().__init__(parent)
        self._read_only = False

    def set_read_only(self, read_only: bool) -> None:
        self._read_only = read_only

    def mimeTypes(self) -> list[str]:
        return ["text/plain"]

    def mimeData(self, indexes: Iterable[QModelIndex]) -> QMimeData:
        items = [self.itemFromIndex(index) for index in indexes]
        data = [self._get_parameters_from_item(item) for item in items]
        serialized = serialization.to_json(data)
        mime_data = QMimeData()
        mime_data.setText(serialized)
        return mime_data

    def dropMimeData(
        self,
        data: QMimeData,
        action: Qt.DropAction,
        row: int,
        column: int,
        parent: AnyModelIndex,
    ) -> bool:
        if self._read_only:
            return False
        json_string = data.text()
        try:
            steps = serialization.from_json(
                json_string,
                list[tuple[DottedVariableName, ParameterNamespace | Expression]],
            )
        except ValueError:
            return False

        new_items = [self._create_item(name, value) for name, value in steps]
        if row == -1:
            row = self.rowCount(parent)
        if not (self.flags(parent) & Qt.ItemFlag.ItemIsDropEnabled):
            return False
        parent_item = (
            self.itemFromIndex(parent) if parent.isValid() else self.invisibleRootItem()
        )
        parent_item.insertRows(row, new_items)
        return True

    def canDropMimeData(self, data, action, row, column, parent):
        if self._read_only:
            return False
        return bool(self.flags(parent) & Qt.ItemFlag.ItemIsDropEnabled)

    def set_parameters(self, parameters: ParameterNamespace) -> None:
        self.removeRows(0, self.rowCount(), QModelIndex())
        for name, value in parameters.items():
            item = self._create_item(name, value)
            self.appendRow(item)
        self.modelReset.emit()

    def get_parameters(self) -> ParameterNamespace:
        namespace = []
        root = self.invisibleRootItem()
        for row in range(root.rowCount()):
            item = root.child(row)
            name, value = self._get_parameters_from_item(item)
            namespace.append((name, value))
        return ParameterNamespace(namespace)

    def flags(self, index):
        flags = super().flags(index)
        if self._read_only:
            flags &= ~Qt.ItemFlag.ItemIsEditable
            flags &= ~Qt.ItemFlag.ItemIsDropEnabled
            flags &= ~Qt.ItemFlag.ItemIsDragEnabled
        return flags

    def add_parameter(self, name: DottedVariableName, value: Expression) -> None:
        root = self.invisibleRootItem()
        item = self._create_item(name, value)
        root.appendRow(item)

    def add_namespace(self, name: DottedVariableName) -> None:
        root = self.invisibleRootItem()
        item = self._create_item(name, ParameterNamespace.empty())
        root.appendRow(item)

    def hasChildren(self, parent: AnyModelIndex = DEFAULT_INDEX) -> bool:
        # hasChildren is used to know when to display a new column in the ColumnView,
        # so we only return true when the parent is a namespace.
        if not parent.isValid():
            return True
        item = self.itemFromIndex(parent)
        data = item.data(PARAMETER_VALUE_ROLE)
        assert isinstance(data, Expression) or data is None
        return data is None

    def _get_parameters_from_item(
        self, item: QStandardItem
    ) -> tuple[DottedVariableName, ParameterNamespace | Expression]:
        name = item.data(PARAMETER_NAME_ROLE)
        assert isinstance(name, DottedVariableName)
        value = item.data(PARAMETER_VALUE_ROLE)
        assert isinstance(value, Expression) or value is None
        if value is None:
            namespace = []
            for row in range(item.rowCount()):
                sub_item = item.child(row)
                sub_name, sub_value = self._get_parameters_from_item(sub_item)
                namespace.append((sub_name, sub_value))
            return name, ParameterNamespace(namespace)
        else:
            return name, value

    def _create_item(
        self, name: DottedVariableName, value: ParameterNamespace | Expression
    ) -> QStandardItem:
        item = QStandardItem()
        flags = (
            Qt.ItemFlag.ItemIsEnabled
            | Qt.ItemFlag.ItemIsSelectable
            | Qt.ItemFlag.ItemIsEditable
            | Qt.ItemFlag.ItemIsDragEnabled
        )
        if isinstance(value, Expression):
            item.setData(name, PARAMETER_NAME_ROLE)
            item.setData(value, PARAMETER_VALUE_ROLE)
            flags |= Qt.ItemFlag.ItemNeverHasChildren
        elif isinstance(value, ParameterNamespace):
            item.setData(name, PARAMETER_NAME_ROLE)
            item.setData(None, PARAMETER_VALUE_ROLE)
            flags |= Qt.ItemFlag.ItemIsDropEnabled
            for sub_name, sub_value in value.items():
                sub_item = self._create_item(sub_name, sub_value)
                item.appendRow(sub_item)
            item.setData(None, Qt.ItemDataRole.UserRole)
        else:
            raise ValueError(f"Invalid value {value}")
        item.setFlags(flags)
        return item

    def supportedDropActions(self) -> Qt.DropAction:
        return Qt.DropAction.MoveAction


NAME_COLOR = "#AA4926"
VALUE_COLOR = "#6897BB"


class ParameterEditorDelegate(HTMLItemDelegate):
    """A custom delegate to display and edit the parameters in the view."""

    def get_text_to_render(self, index: QModelIndex) -> str:
        palette = QApplication.palette()
        text_color = f"#{palette.text().color().rgba():X}"
        name = index.data(PARAMETER_NAME_ROLE)
        assert isinstance(name, DottedVariableName)
        value = index.data(PARAMETER_VALUE_ROLE)
        assert isinstance(value, Expression) or value is None

        if value is None:
            return f"<span style='color:{NAME_COLOR}'>{name}</span>"
        else:
            return (
                f"<span style='color:{NAME_COLOR}'>{name}</span> "
                f"<span style='color:{text_color}'>=</span> "
                f"<span style='color:{VALUE_COLOR}'>{value}</span>"
            )

    def createEditor(
        self, parent: QWidget, option: QStyleOptionViewItem, index: AnyModelIndex
    ) -> QWidget:
        name = index.data(PARAMETER_NAME_ROLE)
        assert isinstance(name, DottedVariableName)
        value = index.data(PARAMETER_VALUE_ROLE)
        assert isinstance(value, Expression) or value is None

        if value is None:
            editor = NamespaceEditor(option.font)  # type: ignore[reportAttributeAccessIssue]
        else:
            editor = ParameterEditor(option.font)  # type: ignore[reportAttributeAccessIssue]
        editor.setParent(parent)
        return editor

    def setEditorData(self, editor, index):
        name = index.data(PARAMETER_NAME_ROLE)
        assert isinstance(name, DottedVariableName)
        value = index.data(PARAMETER_VALUE_ROLE)
        assert isinstance(value, Expression) or value is None

        if value is None:
            assert isinstance(editor, NamespaceEditor)
            editor.set_namespace(name)
        else:
            assert isinstance(editor, ParameterEditor)
            editor.set_parameter(name, value)

    def setModelData(self, editor, model, index):
        name = index.data(PARAMETER_NAME_ROLE)
        assert isinstance(name, DottedVariableName)
        value = index.data(PARAMETER_VALUE_ROLE)
        assert isinstance(value, Expression) or value is None

        if value is None:
            assert isinstance(editor, NamespaceEditor)
            new_name = editor.get_namespace()
            model.setData(index, new_name, PARAMETER_NAME_ROLE)
        else:
            assert isinstance(editor, ParameterEditor)
            new_name, new_value = editor.get_parameter()
            model.setData(index, new_name, PARAMETER_NAME_ROLE)
            model.setData(index, new_value, PARAMETER_VALUE_ROLE)


# The editors below have an event filter to remove the focus from them when one of their
# children loses the focus.
# If we don't do this, the editor will keep the focus when the user clicks
# outside of it, and the delegate will not be able to close the editor.
class ParameterEditor(QWidget):
    def __init__(self, font):
        super().__init__()
        self.setFont(font)
        layout = QHBoxLayout(self)
        self.setLayout(layout)
        self.name_editor = AutoResizeLineEdit(self)
        layout.addWidget(self.name_editor)
        label = QLabel("=", self)
        label.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground, True)
        layout.addWidget(label)
        self.value_editor = AutoResizeLineEdit(self)
        layout.addWidget(self.value_editor)
        layout.addStretch(1)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(3)
        palette = self.palette()
        palette.setColor(QPalette.ColorRole.Window, Qt.GlobalColor.black)
        self.setAutoFillBackground(True)
        self.setPalette(palette)
        self.name_editor.setStyleSheet(f"color: {NAME_COLOR}")
        self.value_editor.setStyleSheet(f"color: {VALUE_COLOR}")
        self.name_editor.setPlaceholderText("Parameter name")
        self.value_editor.setPlaceholderText("Parameter value")
        self.name_editor.installEventFilter(self)
        self.value_editor.installEventFilter(self)

    def set_parameter(self, name: DottedVariableName, value: Expression) -> None:
        self.name_editor.setText(str(name))
        self.value_editor.setText(str(value))

    def get_parameter(self) -> tuple[DottedVariableName, Expression]:
        name = DottedVariableName(self.name_editor.text())
        value = Expression(self.value_editor.text())
        return name, value

    def eventFilter(self, watched: QObject, event: QEvent) -> bool:
        watched_is_child = watched is self.name_editor or watched is self.value_editor
        if watched_is_child and event.type() == QEvent.Type.FocusOut:
            if QApplication.focusWidget() not in self.findChildren(QWidget):
                QApplication.postEvent(self, QFocusEvent(event.type()))
            return False

        return super().eventFilter(watched, event)


class NamespaceEditor(QWidget):
    def __init__(self, font):
        super().__init__()
        self.setFont(font)
        layout = QVBoxLayout(self)
        self.setLayout(layout)
        self.name_editor = AutoResizeLineEdit(self)
        layout.addWidget(self.name_editor)
        layout.addStretch(1)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        palette = self.palette()
        palette.setColor(QPalette.ColorRole.Window, Qt.GlobalColor.black)
        self.setAutoFillBackground(True)
        self.setPalette(palette)
        self.name_editor.setStyleSheet(f"color: {NAME_COLOR}")
        self.name_editor.setPlaceholderText("Namespace name")
        self.name_editor.installEventFilter(self)

    def set_namespace(self, name: DottedVariableName) -> None:
        self.name_editor.setText(str(name))

    def get_namespace(self) -> DottedVariableName:
        name = DottedVariableName(self.name_editor.text())
        return name

    def eventFilter(self, watched: QObject, event: QEvent) -> bool:
        if watched is self.name_editor and event.type() == QEvent.Type.FocusOut:
            QApplication.postEvent(self, QFocusEvent(event.type()))
            return False

        return super().eventFilter(watched, event)



from caqtus.gui.condetrol.timelanes_editor import TimeLaneModel
from caqtus.gui.condetrol.timelanes_editor.analog_lane_editor import AnalogTimeLaneModel
from caqtus.types.expression import Expression
from caqtus.types.timelane import AnalogTimeLane
from caqtus.utils import serialization
from caqtus.utils.serialization import JSON
from ._extension import TimeLaneExtension


def create_analog_lane(number_steps: int) -> AnalogTimeLane:
    return AnalogTimeLane([Expression("...")] * number_steps)


def create_lane_model(lane, lane_name: str) -> TimeLaneModel:
    model = AnalogTimeLaneModel(lane_name)
    model.set_lane(lane)
    return model


def create_lane_delegate(lane, lane_name: str) -> None:
    return None


def dump_analog_lane(time_lane: AnalogTimeLane):
    return serialization.converters["json"].unstructure(time_lane, AnalogTimeLane)


def load_analog_lane(content: JSON):
    return serialization.converters["json"].structure(content, AnalogTimeLane)


analog_time_lane_extension = TimeLaneExtension(
    label="Analog",
    lane_type=AnalogTimeLane,
    dumper=dump_analog_lane,
    loader=load_analog_lane,
    lane_factory=create_analog_lane,
    lane_model_factory=create_lane_model,
    lane_delegate_factory=create_lane_delegate,
    type_tag="analog",
)



from __future__ import annotations

import abc
import contextlib
import datetime
import json
import weakref
from collections.abc import Awaitable
from typing import Optional, TypeGuard, assert_never, assert_type, Never, Callable

import anyio
import anyio.abc
import anyio.lowlevel
import attrs
from PySide6.QtCore import (
    QObject,
    QAbstractItemModel,
    QModelIndex,
    Qt,
    QDateTime,
    QPersistentModelIndex,
    QMimeData,
)
from PySide6.QtGui import QStandardItemModel, QStandardItem

from caqtus.session import (
    ExperimentSessionMaker,
    PureSequencePath,
    ExperimentSession,
    AsyncExperimentSession,
    PathIsRootError,
    PathHasChildrenError,
)
from caqtus.session import (
    PathNotFoundError,
    PathIsSequenceError,
    PathIsNotSequenceError,
    State,
)
from caqtus.session._sequence_collection import SequenceStats
from caqtus.types.iteration import (
    Unknown,
    IterationConfiguration,
)
from caqtus.types.timelane import TimeLanes
from caqtus.utils.result import (
    is_failure,
    Failure,
    Success,
    is_failure_type,
    is_success,
)

NODE_DATA_ROLE = Qt.ItemDataRole.UserRole + 1

DEFAULT_INDEX = QModelIndex()


def get_item_data(item: QStandardItem) -> Node:
    data = item.data(NODE_DATA_ROLE)
    assert is_node(data)
    return data


@attrs.define
class FolderNode:
    path: PureSequencePath
    has_fetched_children: bool
    creation_date: datetime.datetime


@attrs.define
class SequenceNode(abc.ABC):
    path: PureSequencePath
    stats: SequenceStats
    creation_date: datetime.datetime
    last_query_time: datetime.datetime


Node = FolderNode | SequenceNode


def is_node(value) -> TypeGuard[Node]:
    return isinstance(value, (FolderNode, SequenceNode))


class AsyncPathHierarchyModel(QAbstractItemModel):
    def __init__(
        self, session_maker: ExperimentSessionMaker, parent: Optional[QObject] = None
    ):
        super().__init__(parent)
        self.session_maker = session_maker

        self.tree = QStandardItemModel(self)
        self.tree.invisibleRootItem().setData(
            FolderNode(
                path=PureSequencePath.root(),
                has_fetched_children=False,
                creation_date=datetime.datetime.min,
            ),
            NODE_DATA_ROLE,
        )
        self._background_runner = BackgroundRunner(self.watch_session)

    def index(self, row, column, parent=DEFAULT_INDEX):
        if not self.hasIndex(row, column, parent):
            return QModelIndex()
        parent_item = (
            parent.internalPointer()
            if parent.isValid()
            else self.tree.invisibleRootItem()
        )
        child_item = parent_item.child(row)
        return (
            self.createIndex(row, column, child_item) if child_item else QModelIndex()
        )

    def parent(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, index: QModelIndex | QPersistentModelIndex = DEFAULT_INDEX
    ):
        if not index.isValid():
            return QModelIndex()
        child_item = index.internalPointer()
        parent_item = child_item.parent()
        if parent_item is None:
            return QModelIndex()
        return (
            self.createIndex(parent_item.row(), 0, parent_item)
            if parent_item is not self.tree.invisibleRootItem()
            else QModelIndex()
        )

    def flags(self, index: QModelIndex | QPersistentModelIndex) -> Qt.ItemFlag:
        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags | Qt.ItemFlag.ItemIsDropEnabled
        item = self._get_item(index)
        node_data = get_item_data(item)
        flags = Qt.ItemFlag.NoItemFlags
        if isinstance(node_data, SequenceNode):
            flags |= Qt.ItemFlag.ItemIsEnabled
            flags |= Qt.ItemFlag.ItemIsSelectable
            flags |= Qt.ItemFlag.ItemNeverHasChildren
            if node_data.stats.state not in {State.PREPARING, State.RUNNING}:
                flags |= Qt.ItemFlag.ItemIsDragEnabled
        else:
            assert_type(node_data, FolderNode)
            if index.column() == 0:
                flags |= Qt.ItemFlag.ItemIsSelectable
                flags |= Qt.ItemFlag.ItemIsEnabled
                flags |= Qt.ItemFlag.ItemIsDragEnabled
                flags |= Qt.ItemFlag.ItemIsDropEnabled
            if index.column() == 4:
                flags |= Qt.ItemFlag.ItemIsSelectable
                flags |= Qt.ItemFlag.ItemIsEnabled
        return flags

    def mimeTypes(self):  # noqa: N802
        return ["application/x-caqtus-sequence-path"]

    def mimeData(self, indexes):  # noqa: N802
        items = [self._get_item(index) for index in indexes]
        data = [get_item_data(item) for item in items]
        paths = {node.path for node in data}
        assert len(paths) == 1
        encoded = json.dumps([str(path) for path in paths]).encode("utf-8")
        mime_data = QMimeData()
        mime_data.setData("application/x-caqtus-sequence-path", encoded)
        return mime_data

    def supportedDropActions(self):  # noqa: N802
        return Qt.DropAction.MoveAction

    def canDropMimeData(self, data, action, row, column, parent):  # noqa: N802
        if not data.hasFormat("application/x-caqtus-sequence-path"):
            return False
        if action != Qt.DropAction.MoveAction:
            return False
        parent_item = self._get_item(parent)
        node_data = get_item_data(parent_item)
        if row == -1:
            return isinstance(node_data, FolderNode)
        return False

    def dropMimeData(self, data, action, row, column, parent):  # noqa: N802
        if not self.canDropMimeData(data, action, row, column, parent):
            return False

        paths = json.loads(
            bytes(
                data.data(
                    "application/x-caqtus-sequence-path"
                )  # pyright: ignore[reportArgumentType]
            ).decode("utf-8")
        )
        assert len(paths) == 1
        src = PureSequencePath(paths[0])
        if row == -1:
            dst_parent = get_item_data(self._get_item(parent)).path
            dst = dst_parent / src.name
            with self.session_maker() as session, self._background_runner.suspend():
                result = session.paths.move(src, dst)
                return is_success(result)
        return False

    def removeRows(self, row, count, parent=...):  # noqa: N802
        # This method only remove the in-memory data, it does not remove the path from
        # the session.
        parent_item = self._get_item(parent)
        with self._background_runner.suspend():
            self.beginRemoveRows(parent, row, row + count - 1)
            parent_item.removeRows(row, count)
            self.endRemoveRows()
        return True

    def _get_item(self, index) -> QStandardItem:
        result = (
            index.internalPointer()
            if index.isValid()
            else self.tree.invisibleRootItem()
        )
        assert isinstance(result, QStandardItem)
        return result

    def rowCount(self, parent=DEFAULT_INDEX) -> int:  # noqa: N802
        if parent.column() > 0:
            return 0
        parent_item = self._get_item(parent)
        node_data = get_item_data(parent_item)
        if isinstance(node_data, SequenceNode):
            return 0
        else:
            assert_type(node_data, FolderNode)
            if node_data.has_fetched_children:
                return parent_item.rowCount()
            else:
                return 0

    def hasChildren(self, parent=DEFAULT_INDEX) -> bool:  # noqa: N802
        parent_item = self._get_item(parent)
        node_data = get_item_data(parent_item)
        if isinstance(node_data, SequenceNode):
            return False
        else:
            assert_type(node_data, FolderNode)
            if node_data.has_fetched_children:
                return parent_item.rowCount() > 0
            else:
                return True

    def canFetchMore(self, parent) -> bool:  # noqa: N802
        parent_item = self._get_item(parent)
        node_data = get_item_data(parent_item)
        match node_data:
            case SequenceNode():
                return False
            case FolderNode(has_fetched_children=already_fetched):
                return not already_fetched

    def fetchMore(self, parent):  # noqa: N802
        parent_item = self._get_item(parent)
        node_data = get_item_data(parent_item)
        match node_data:
            case SequenceNode():
                return
            case FolderNode(has_fetched_children=True):
                return
            case FolderNode(path=parent_path, has_fetched_children=False):
                assert parent_item.rowCount() == 0
                with self.session_maker() as session:
                    children_result = session.paths.get_children(parent_path)
                    if is_failure_type(children_result, PathIsSequenceError):
                        self.handle_folder_became_sequence(parent, session)
                        return
                    elif is_failure_type(children_result, PathNotFoundError):
                        self.handle_path_was_deleted(parent)
                        return
                    children = children_result.value
                    self.beginInsertRows(parent, 0, len(children) - 1)
                    for child_path in children:
                        child_item = self._build_item(child_path, session)
                        parent_item.appendRow(child_item)
                    node_data.has_fetched_children = True
                    self.endInsertRows()

    @staticmethod
    def _build_item(
        path: PureSequencePath, session: ExperimentSession
    ) -> QStandardItem:
        assert session.paths.does_path_exists(path)
        item = QStandardItem()
        item.setData(path.name, Qt.ItemDataRole.DisplayRole)
        is_sequence_result = session.sequences.is_sequence(path)
        assert not is_failure_type(is_sequence_result, PathNotFoundError)
        is_sequence = is_sequence_result.value
        creation_date_result = session.paths.get_path_creation_date(path)
        assert not is_failure_type(creation_date_result, PathNotFoundError)
        assert not is_failure_type(creation_date_result, PathIsRootError)
        creation_date = creation_date_result.value
        if is_sequence:
            stats_result = session.sequences.get_stats(path)
            assert not is_failure_type(stats_result, PathNotFoundError)
            assert not is_failure_type(stats_result, PathIsNotSequenceError)
            stats = stats_result.value
            item.setData(
                SequenceNode(
                    path=path,
                    stats=stats,
                    creation_date=creation_date,
                    last_query_time=get_update_date(),
                ),
                NODE_DATA_ROLE,
            )
        else:
            item.setData(
                FolderNode(
                    path=path, has_fetched_children=False, creation_date=creation_date
                ),
                NODE_DATA_ROLE,
            )
        return item

    @staticmethod
    async def _build_item_async(
        path: PureSequencePath, session: AsyncExperimentSession
    ) -> QStandardItem:
        assert await session.paths.does_path_exists(path)
        item = QStandardItem()
        item.setData(path.name, Qt.ItemDataRole.DisplayRole)
        is_sequence_result = await session.sequences.is_sequence(path)
        assert not is_failure_type(is_sequence_result, PathNotFoundError)
        is_sequence = is_sequence_result.value
        creation_date_result = await session.paths.get_path_creation_date(path)
        assert not is_failure_type(creation_date_result, PathNotFoundError)
        assert not is_failure_type(creation_date_result, PathIsRootError)
        creation_date = creation_date_result.value
        if is_sequence:
            stats_result = await session.sequences.get_stats(path)
            assert not is_failure_type(stats_result, PathNotFoundError)
            assert not is_failure_type(stats_result, PathIsNotSequenceError)
            stats = stats_result.value
            item.setData(
                SequenceNode(
                    path=path,
                    stats=stats,
                    creation_date=creation_date,
                    last_query_time=get_update_date(),
                ),
                NODE_DATA_ROLE,
            )
        else:
            item.setData(
                FolderNode(
                    path=path, has_fetched_children=False, creation_date=creation_date
                ),
                NODE_DATA_ROLE,
            )
        return item

    def columnCount(self, parent=DEFAULT_INDEX) -> int:  # noqa: N802
        return 5

    def headerData(  # noqa: N802
        self, section, orientation, role=Qt.ItemDataRole.DisplayRole
    ):
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                if section == 0:
                    return "Name"
                elif section == 1:
                    return "Status"
                elif section == 2:
                    return "Progress"
                elif section == 3:
                    return "Duration"
                elif section == 4:
                    return "Date created"
            else:
                return section
        return None

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        """Get the data for a specific index in the model.

        The displayed data returned for each column is as follows:
        0: Name
        A string with the name of the folder or sequence.
        1: Status
        The status of the sequence.
        It is None for folders and a SequenceStats object for sequences.
        2: Progress
        A string representing the number of completed shots and the total
        number of shots of the sequence.
        It is None for folders.
        3: Duration
        A string representing the elapsed and remaining time of the
        sequence.
        It is None for folders.
        4: Date created
        A QDateTime object representing the date and time when the
        sequence or folder was created.
        """

        if not index.isValid() or role != Qt.ItemDataRole.DisplayRole:
            return None
        item = self._get_item(index)
        node_data = get_item_data(item)
        if index.column() == 0:
            return node_data.path.name
        elif index.column() == 1:
            if isinstance(node_data, SequenceNode):
                return node_data.stats
            else:
                return None
        elif index.column() == 2:
            if isinstance(node_data, SequenceNode):
                return (
                    f"{node_data.stats.number_completed_shots}"
                    f"/{node_data.stats.expected_number_shots}"
                )
            else:
                return None
        elif index.column() == 3:
            if isinstance(node_data, SequenceNode):
                return format_duration(node_data.stats, node_data.last_query_time)
            else:
                return None
        elif index.column() == 4:
            return QDateTime(node_data.creation_date.astimezone(None))  # type: ignore

    async def run(self) -> Never:
        await self._background_runner.run()

    async def watch_session(self) -> None:
        while True:
            # The call to update_from_session must be safe to be cancelled in the
            # middle of its execution, without corrupting the model.
            await self.update_from_session()

    async def update_from_session(self) -> None:
        await self.prune()
        await self.add_new_paths()
        for row in range(self.rowCount()):
            await self.update_stats(self.index(row, 0))

    async def update_stats(self, index: QModelIndex) -> None:
        """Update the stats of sequences and folders in the model from the session."""

        if not index.isValid():
            # This situation occurs sometimes, but unsure why.
            # Maybe if fetch is called in the middle of an async call?
            return None

        item = self._get_item(index)
        data = get_item_data(item)
        change_detected = False
        async with self.session_maker.async_session() as session:
            creation_date_result = await session.paths.get_path_creation_date(data.path)
            assert not is_failure_type(creation_date_result, PathIsRootError)
            if is_failure_type(creation_date_result, PathNotFoundError):
                await self.handle_path_was_deleted_async(index)
                return
            creation_date = creation_date_result.value
            if creation_date != data.creation_date:
                await anyio.lowlevel.checkpoint()
                data.creation_date = creation_date
                change_detected = True
            if isinstance(data, SequenceNode):
                sequence_stats_result = await session.sequences.get_stats(data.path)
                assert not is_failure_type(sequence_stats_result, PathNotFoundError)
                if is_failure_type(sequence_stats_result, PathIsNotSequenceError):
                    await self.handle_sequence_became_folder(index, session)
                    return
                stats = sequence_stats_result.value
                if stats != data.stats:
                    await anyio.lowlevel.checkpoint()
                    data.stats = stats
                    data.last_query_time = get_update_date()
                    change_detected = True
        if change_detected:
            top_left = index.siblingAtColumn(0)
            bottom_right = index.siblingAtColumn(self.columnCount() - 1)
            self.dataChanged.emit(top_left, bottom_right, [Qt.ItemDataRole.DisplayRole])
        if isinstance(data, FolderNode):
            for row in range(item.rowCount()):
                await self.update_stats(self.index(row, 0, index))

    async def prune(self, parent: QModelIndex = DEFAULT_INDEX) -> None:
        """Removes children of the parent that are no longer present in the session."""

        parent_item = self._get_item(parent)
        parent_data = get_item_data(parent_item)

        if isinstance(parent_data, SequenceNode):
            return

        async with self.session_maker.async_session() as session:
            children_result = await session.paths.get_children(parent_data.path)
            await anyio.lowlevel.checkpoint()
            if is_failure_type(children_result, PathIsSequenceError):
                await self.handle_folder_became_sequence_async(parent, session)
                return
            elif is_failure_type(children_result, PathNotFoundError):
                await self.handle_path_was_deleted_async(parent)
                return
            child_paths = children_result.value

        await anyio.lowlevel.checkpoint()
        # Need to use persistent indices to avoid invalidation while removing rows.
        child_indices = set[QPersistentModelIndex]()
        for row in range(self.rowCount(parent)):
            child_indices.add(QPersistentModelIndex(self.index(row, 0, parent)))

        remaining_children = set()
        for child in child_indices:
            child_item = self._get_item(child)
            child_path = get_item_data(child_item).path
            if child_path not in child_paths:
                self.beginRemoveRows(parent, child.row(), child.row())
                parent_item.removeRow(child.row())
                self.endRemoveRows()
            else:
                remaining_children.add(child)

        for child in remaining_children:
            await self.prune(QModelIndex(child))

    async def add_new_paths(self, parent: QModelIndex = DEFAULT_INDEX) -> None:
        """Add new paths to the model that have been added to the session."""

        parent_item = self._get_item(parent)
        parent_data = get_item_data(parent_item)
        match parent_data:
            case SequenceNode():
                return
            case FolderNode(has_fetched_children=False):
                return
            case FolderNode(path=parent_path, has_fetched_children=True):
                async with self.session_maker.async_session() as session:
                    children_result = await session.paths.get_children(parent_path)
                    if is_failure_type(children_result, PathIsSequenceError):
                        await self.handle_folder_became_sequence_async(parent, session)
                        return
                    elif is_failure_type(children_result, PathNotFoundError):
                        await self.handle_path_was_deleted_async(parent)
                        return
                    child_paths = children_result.value
                    already_added_paths = {
                        get_item_data(parent_item.child(row)).path
                        for row in range(parent_item.rowCount())
                    }
                    new_paths = child_paths - already_added_paths
                    new_items = [
                        await self._build_item_async(path, session)
                        for path in new_paths
                    ]
                    await self.append_items(parent, new_items)
                for row in range(self.rowCount(parent)):
                    await self.add_new_paths(self.index(row, 0, parent))

    async def append_items(
        self, parent: QModelIndex, items: list[QStandardItem]
    ) -> None:
        await anyio.lowlevel.checkpoint()
        parent_item = self._get_item(parent)
        self.beginInsertRows(
            parent,
            parent_item.rowCount(),
            parent_item.rowCount() + len(items) - 1,
        )
        for item in items:
            parent_item.appendRow(item)
        self.endInsertRows()

    def handle_folder_became_sequence(
        self, index: QModelIndex | QPersistentModelIndex, session: ExperimentSession
    ):
        item = self._get_item(index)
        data = get_item_data(item)
        stats_result = session.sequences.get_stats(data.path)
        assert not is_failure_type(stats_result, PathNotFoundError)
        assert not is_failure_type(stats_result, PathIsNotSequenceError)
        stats = stats_result.value
        creation_date_result = session.paths.get_path_creation_date(data.path)
        assert not is_failure_type(creation_date_result, PathNotFoundError)
        assert not is_failure_type(creation_date_result, PathIsRootError)
        creation_date = creation_date_result.value
        self.beginRemoveRows(index, 0, item.rowCount() - 1)
        item.setData(
            SequenceNode(
                path=data.path,
                stats=stats,
                creation_date=creation_date,
                last_query_time=get_update_date(),
            ),
            NODE_DATA_ROLE,
        )
        item.removeRows(0, item.rowCount())
        self.endRemoveRows()
        self.emit_index_updated(index)

    async def handle_folder_became_sequence_async(
        self, index: QModelIndex, session: AsyncExperimentSession
    ):
        item = self._get_item(index)
        data = get_item_data(item)
        stats_result = await session.sequences.get_stats(data.path)
        assert not is_failure_type(stats_result, PathNotFoundError)
        assert not is_failure_type(stats_result, PathIsNotSequenceError)
        stats = stats_result.value
        creation_date_result = await session.paths.get_path_creation_date(data.path)
        assert not is_failure_type(creation_date_result, PathNotFoundError)
        assert not is_failure_type(creation_date_result, PathIsRootError)
        creation_date = creation_date_result.value
        await anyio.lowlevel.checkpoint()
        self.beginRemoveRows(index, 0, item.rowCount() - 1)
        item.setData(
            SequenceNode(
                path=data.path,
                stats=stats,
                creation_date=creation_date,
                last_query_time=get_update_date(),
            ),
            NODE_DATA_ROLE,
        )
        item.removeRows(0, item.rowCount())
        self.endRemoveRows()
        self.emit_index_updated(index)

    def handle_path_was_deleted(
        self, index: QModelIndex | QPersistentModelIndex
    ) -> None:
        parent = self.parent(index)
        parent_item = self._get_item(parent)
        self.beginRemoveRows(parent, index.row(), index.row())
        parent_item.removeRow(index.row())
        self.endRemoveRows()

    async def handle_path_was_deleted_async(self, index: QModelIndex) -> None:
        await anyio.lowlevel.checkpoint()
        self.handle_path_was_deleted(index)

    def get_path(self, index: QModelIndex | QPersistentModelIndex) -> PureSequencePath:
        return get_item_data(self._get_item(index)).path

    async def handle_sequence_became_folder(
        self, index: QModelIndex, session: AsyncExperimentSession
    ):
        item = self._get_item(index)
        data = get_item_data(item)
        creation_date_result = await session.paths.get_path_creation_date(data.path)
        assert not is_failure_type(creation_date_result, PathNotFoundError)
        assert not is_failure_type(creation_date_result, PathIsRootError)
        creation_date = creation_date_result.value
        assert item.rowCount() == 0
        await anyio.lowlevel.checkpoint()
        item.setData(
            FolderNode(
                path=data.path, has_fetched_children=False, creation_date=creation_date
            ),
            NODE_DATA_ROLE,
        )
        self.emit_index_updated(index)

    def emit_index_updated(self, index: QModelIndex | QPersistentModelIndex) -> None:
        index = QModelIndex(index)
        self.dataChanged.emit(
            index.siblingAtColumn(0),
            index.siblingAtColumn(self.columnCount() - 1),
            [Qt.ItemDataRole.DisplayRole],
        )

    def rename(
        self, index: QModelIndex, new_name: str
    ) -> Success[None] | Failure[Exception]:
        """Rename a sequence or folder in the model.

        Args:
            index: Index of the sequence or folder to rename.
                It must be a valid index.
            new_name: The new name to give to the path.
                It must be a valid path name.
        """

        if not index.isValid():
            raise ValueError("Invalid index")

        item = self._get_item(index)
        data = get_item_data(item)

        # Since the index is valid, it can't be the root item, and thus it must have a
        # parent.
        assert data.path.parent is not None
        new_path = data.path.parent / new_name

        with self.session_maker() as session, self._background_runner.suspend():
            result = session.paths.move(data.path, new_path)
            if is_success(result):
                self._rename_recursively(index, new_path)
            return result

    def _rename_recursively(
        self, index: QModelIndex, new_prefix: PureSequencePath
    ) -> None:
        item = self._get_item(index)
        data = get_item_data(item)
        new_parts = new_prefix.parts + data.path.parts[len(new_prefix.parts) :]
        new_path = PureSequencePath.from_parts(new_parts)
        data.path = new_path
        self.emit_index_updated(index)
        for row in range(self.rowCount(index)):
            self._rename_recursively(self.index(row, 0, index), new_path)

    def create_new_sequence(
        self,
        parent: QModelIndex,
        name: str,
        iteration_config: IterationConfiguration,
        time_lanes: TimeLanes,
    ) -> Success[None] | Failure[PathIsSequenceError] | Failure[PathHasChildrenError]:
        parent_item = self._get_item(parent)
        parent_data = get_item_data(parent_item)
        if not isinstance(parent_data, FolderNode):
            raise ValueError("Parent must be a folder")
        new_path = parent_data.path / name
        with self.session_maker() as session, self._background_runner.suspend():
            creation_result = session.sequences.create(
                new_path, iteration_config, time_lanes
            )
            if is_failure(creation_result):
                return creation_result
            item = self._build_item(new_path, session)
            self.beginInsertRows(parent, parent_item.rowCount(), parent_item.rowCount())
            parent_item.appendRow(item)
            self.endInsertRows()
            return creation_result

    def remove_path(
        self, index: QModelIndex
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathIsSequenceError]
        | Failure[PathIsRootError]
    ):
        """Remove the path at the requested index."""

        item = self._get_item(index)
        path = get_item_data(item).path
        with self.session_maker() as session:
            is_sequence_result = session.sequences.is_sequence(path)
            if is_failure(is_sequence_result):
                return is_sequence_result
            with self._background_runner.suspend():
                if is_sequence_result.value:
                    result = session.paths.delete_path(path, delete_sequences=True)
                else:
                    result = session.paths.delete_path(path, delete_sequences=False)
            if is_success(result):
                assert self.removeRows(index.row(), 1, self.parent(index))
                return Success(None)
            else:
                return result


def format_duration(stats: SequenceStats, updated_time: datetime.datetime) -> str:
    if stats.state == State.DRAFT or stats.state == State.PREPARING:
        return "--/--"
    elif stats.state == State.RUNNING:
        assert stats.start_time is not None
        assert stats.number_completed_shots is not None
        running_duration = updated_time - stats.start_time
        expected_num_shots = stats.expected_number_shots
        if isinstance(expected_num_shots, Unknown) or stats.number_completed_shots == 0:
            remaining = "--"
        else:
            remaining = (
                running_duration
                / stats.number_completed_shots
                * (expected_num_shots - stats.number_completed_shots)
            )
        if isinstance(remaining, datetime.timedelta):
            total = remaining + running_duration
            remaining = _format_seconds(total.total_seconds())
        running_duration = _format_seconds(running_duration.total_seconds())
        return f"{running_duration}/{remaining}"
    elif (
        stats.state == State.FINISHED
        or stats.state == State.CRASHED
        or stats.state == State.INTERRUPTED
    ):
        # Need to handle the case where the sequence crashed before starting.
        if stats.start_time is None or stats.stop_time is None:
            return ""
        else:
            total_duration = stats.stop_time - stats.start_time
            return _format_seconds(total_duration.total_seconds())
    assert_never(stats.state)


def _format_seconds(seconds: float) -> str:
    """Format seconds into a string.

    Args:
        seconds: Seconds to format.

    Returns:
        Formatted string.
    """

    seconds = int(seconds)
    result = [f"{seconds % 60}s"]

    minutes = seconds // 60
    if minutes > 0:
        result.append(f"{minutes % 60}m")
        hours = minutes // 60
        if hours > 0:
            result.append(f"{hours % 24}h")
            days = hours // 24
            if days > 0:
                result.append(f"{days}d")

    return ":".join(reversed(result))


def get_update_date() -> datetime.datetime:
    return datetime.datetime.now(tz=datetime.timezone.utc).replace(microsecond=0)


class BackgroundRunner:
    def __init__(self, task: Callable[[], Awaitable[None]]):
        self.task = task
        self._task_group: Optional[anyio.abc.TaskGroup] = None
        self._cancel_scopes = weakref.WeakValueDictionary[int, anyio.CancelScope]()

    async def run(self) -> Never:
        """Run the background task indefinitely."""

        while True:
            self._can_resume = anyio.Event()
            with anyio.CancelScope() as self._cancel_scope:
                await self.task()
            await self._can_resume.wait()

    @contextlib.contextmanager
    def suspend(self):
        """Suspend the background task until the context manager exits.

        When the context manager is entered, the background task is cancelled at the
        next cancellation point.
        When the context manager exits, a new background task is started.
        """

        self._cancel_scope.cancel()

        yield

        self._can_resume.set()



from __future__ import annotations

import contextlib
import copy
import logging
from collections.abc import Mapping, AsyncGenerator, AsyncIterable
from typing import Optional

import anyio
import anyio.to_process

from caqtus.device import DeviceName, DeviceConfiguration
from caqtus.session import (
    ExperimentSessionMaker,
    PureSequencePath,
    TracebackSummary,
)
from caqtus.session._shot_id import ShotId
from caqtus.shot_compilation import (
    DeviceCompiler,
    SequenceContext,
)
from caqtus.types.iteration import StepsConfiguration
from caqtus.types.parameter import ParameterNamespace
from caqtus.types.recoverable_exceptions import split_recoverable
from caqtus.types.timelane.timelane import TimeLanes
from caqtus.utils.result import unwrap
from caqtus.utils.result._result import is_failure
from ._shot_compiler import ShotCompilerFactory, create_shot_compiler
from ._shot_runner import ShotRunnerFactory, create_shot_runner
from .sequence_runner import execute_steps
from .shots_manager import ShotManager, ShotData, ShotScheduler, ShotRetryConfig
from ..device_manager_extension import DeviceManagerExtensionProtocol
from ...types.iteration._step_context import StepContext

logger = logging.getLogger(__name__)


async def run_sequence(
    sequence: PureSequencePath,
    session_maker: ExperimentSessionMaker,
    shot_retry_config: Optional[ShotRetryConfig],
    global_parameters: Optional[ParameterNamespace],
    device_configurations: Optional[Mapping[DeviceName, DeviceConfiguration]],
    device_manager_extension: DeviceManagerExtensionProtocol,
    shot_runner_factory: ShotRunnerFactory = create_shot_runner,
    shot_compiler_factory: ShotCompilerFactory = create_shot_compiler,
) -> None:
    """Manages the execution of a sequence.

    Args:
        sequence: The sequence to run.

        session_maker: A factory for creating experiment sessions.
            This is used to connect to the storage in which to find the sequence.

        shot_retry_config: Specifies how to retry a shot if an error occurs.
            If an error occurs when the shot runner is running a shot, it will be caught
            by the sequence manager and the shot will be retried according to the
            configuration in this object.

        global_parameters: The global parameters to use to run the sequence.

            These parameters will be saved as the global parameters for the sequence
            when it is prepared.

            If None, the sequence manager will use the default global parameters stored
            in the session.

        device_configurations: The device configurations to use to run the sequence.

            These configurations will be saved as the configurations used for the
            sequence when it is prepared.

            If None, the sequence manager will use the default device configurations.

        device_manager_extension: Used to instantiate the device components.

        shot_runner_factory: A function that can be used to create an object to run
            shots.

        shot_compiler_factory: A function that can be used to create an object to
            compile shots.
    """

    with session_maker.session() as session:
        iteration = session.sequences.get_iteration_configuration(sequence)
    if not isinstance(iteration, StepsConfiguration):
        raise NotImplementedError("Only steps iteration is supported at the moment.")
    sequence_manager = SequenceManager(
        sequence=sequence,
        session_maker=session_maker,
        shot_retry_config=shot_retry_config,
        global_parameters=global_parameters,
        device_configurations=device_configurations,
        device_manager_extension=device_manager_extension,
        shot_runner_factory=shot_runner_factory,
        shot_compiler_factory=shot_compiler_factory,
    )
    initial_context = StepContext(sequence_manager.sequence_parameters.evaluate())
    async with sequence_manager.run_sequence() as shot_scheduler:
        await execute_steps(iteration, initial_context, shot_scheduler)


class SequenceManager:
    def __init__(
        self,
        sequence: PureSequencePath,
        session_maker: ExperimentSessionMaker,
        shot_retry_config: Optional[ShotRetryConfig],
        global_parameters: Optional[ParameterNamespace],
        device_configurations: Optional[Mapping[DeviceName, DeviceConfiguration]],
        device_manager_extension: DeviceManagerExtensionProtocol,
        shot_runner_factory: ShotRunnerFactory,
        shot_compiler_factory: ShotCompilerFactory,
    ) -> None:
        self._session_maker = session_maker
        self._sequence_path = sequence
        self._shot_retry_config = shot_retry_config or ShotRetryConfig()

        with self._session_maker() as session:
            if device_configurations is None:
                self.device_configurations = dict(session.default_device_configurations)
            else:
                self.device_configurations = dict(device_configurations)
            if global_parameters is None:
                self.sequence_parameters = session.get_global_parameters()
            else:
                self.sequence_parameters = copy.deepcopy(global_parameters)
            time_lanes = session.sequences.get_time_lanes(self._sequence_path)
            if is_failure(time_lanes):
                raise time_lanes.exception()
            self.time_lanes: TimeLanes = time_lanes

        self._device_manager_extension = device_manager_extension
        self._device_compilers: dict[DeviceName, DeviceCompiler] = {}

        self._shot_runner_factory = shot_runner_factory
        self._shot_compiler_factory = shot_compiler_factory

    @contextlib.asynccontextmanager
    async def run_sequence(self) -> AsyncGenerator[ShotScheduler, None]:
        """Run background tasks to compile and run shots for a given sequence.

        Returns:
            A asynchronous context manager that yields a shot scheduler object.

            When the context manager is entered, it will set the sequence to PREPARING
            while acquiring the necessary resources and the transition to RUNNING.

            The context manager will yield a shot scheduler object that can be used to
            push shots to the sequence execution queue.
            When a shot is done, its associated data will be stored in the associated
            sequence.

            One shot scheduling is over, the context manager will be exited.
            At this point is will finish the sequence and transition the sequence state
            to FINISHED when the sequence terminated normally, CRASHED if an error
            occurred or INTERRUPTED if the sequence was interrupted by the user.
        """

        with self._session_maker() as session:
            session.sequences.set_preparing(
                self._sequence_path,
                self.device_configurations,
                self.sequence_parameters,
            )
        try:
            sequence_context = SequenceContext(
                device_configurations=self.device_configurations,  # pyright: ignore[reportCallIssue]
                time_lanes=self.time_lanes,  # pyright: ignore[reportCallIssue]
            )
            shot_compiler = self._shot_compiler_factory(
                sequence_context,
                self._device_manager_extension,
            )

            # We start the subprocesses while preparing the sequence to avoid
            # the overhead of starting when the sequence is launched.
            await anyio.to_process.run_sync(nothing)
            async with (
                self._shot_runner_factory(
                    sequence_context, shot_compiler, self._device_manager_extension
                ) as shot_runner,
                ShotManager(
                    shot_runner,
                    shot_compiler,
                    self._shot_retry_config,
                ) as (
                    scheduler_cm,
                    data_stream_cm,
                ),
            ):
                with self._session_maker() as session:
                    session.sequences.set_running(self._sequence_path, start_time="now")
                async with (
                    anyio.create_task_group() as tg,
                    scheduler_cm as scheduler,
                ):
                    tg.start_soon(self._store_shots, data_stream_cm)
                    yield scheduler
        except* anyio.get_cancelled_exc_class():
            with self._session_maker() as session:
                session.sequences.set_interrupted(self._sequence_path, stop_time="now")
            raise
        except* BaseException as e:
            tb_summary = TracebackSummary.from_exception(e)
            with self._session_maker() as session:
                unwrap(
                    session.sequences.set_crashed(
                        self._sequence_path, tb_summary, stop_time="now"
                    )
                )
            recoverable, non_recoverable = split_recoverable(e)
            if non_recoverable:
                raise
            if recoverable:
                logger.warning(
                    "A recoverable error occurred while running the sequence.",
                    exc_info=recoverable,
                )

        else:
            with self._session_maker() as session:
                session.sequences.set_finished(self._sequence_path, stop_time="now")

    async def _store_shots(
        self,
        data_stream_cm: contextlib.AbstractAsyncContextManager[AsyncIterable[ShotData]],
    ):
        async with data_stream_cm as shots_data:
            async for shot_data in shots_data:
                await self._store_shot(shot_data)

    async def _store_shot(self, shot_data: ShotData) -> None:
        params = {
            name: value for name, value in shot_data.variables.to_flat_dict().items()
        }
        async with self._session_maker.async_session() as session:
            result = await session.sequences.create_shot(
                ShotId(self._sequence_path, shot_data.index),
                params,
                shot_data.data,
                shot_data.start_time,
                shot_data.end_time,
            )
            unwrap(result)


def nothing():
    pass



from __future__ import annotations

from abc import abstractmethod
from datetime import datetime
from typing import Protocol, TYPE_CHECKING

from caqtus.utils.result import Success, Failure
from .._path import PureSequencePath
from .._path_hierarchy import PathNotFoundError, PathIsRootError

if TYPE_CHECKING:
    from .._sequence_collection import PathIsSequenceError


class AsyncPathHierarchy(Protocol):
    @abstractmethod
    async def does_path_exists(self, path: PureSequencePath) -> bool:
        raise NotImplementedError

    @abstractmethod
    async def get_children(
        self, path: PureSequencePath
    ) -> (
        Success[set[PureSequencePath]]
        | Failure[PathNotFoundError]
        | Failure[PathIsSequenceError]
    ):
        raise NotImplementedError

    @abstractmethod
    async def get_path_creation_date(
        self, path: PureSequencePath
    ) -> Success[datetime] | Failure[PathNotFoundError] | Failure[PathIsRootError]:
        raise NotImplementedError



import contextlib
from collections.abc import AsyncGenerator, Generator
from typing import Protocol


class Closeable(Protocol):
    def close(self): ...


@contextlib.contextmanager
def close_on_error[R: Closeable](resource: R) -> Generator[R, None, None]:
    """Context manager that closes a resource if an error occurs.

    Beware that the resource will NOT be closed if the context manager is exited
    without an exception.
    """

    try:
        yield resource
    except:
        resource.close()
        raise


class AsyncCloseable(Protocol):
    async def aclose(self): ...


@contextlib.asynccontextmanager
async def aclose_on_error[R: AsyncCloseable](resource: R) -> AsyncGenerator[R, None]:
    """Async context manager that closes a resource if an error occurs.

    Beware that the resource will NOT be closed if the context manager is exited
    without an exception.
    """

    try:
        yield resource
    except:
        await resource.aclose()
        raise



"""This package defines widgets that are used to edit device configurations.

All editors must derive from :class:`DeviceConfigurationEditor`.
"""

from ._device_configuration_editor import (
    DeviceConfigurationEditor,
    FormDeviceConfigurationEditor,
)

__all__ = [
    "DeviceConfigurationEditor",
    "FormDeviceConfigurationEditor",
]



#!/usr/bin/python
from PySide6 import QtWidgets

from .custom_widget_value_edit import _NumberValueEdit
from .prop_widgets_abstract import BaseProperty


class _PropVector(BaseProperty):
    """
    Base widget for the PropVector widgets.
    """

    def __init__(self, parent=None, fields=0):
        super(_PropVector, self).__init__(parent)
        self._value = []
        self._items = []
        self._can_emit = True

        layout = QtWidgets.QHBoxLayout(self)
        layout.setSpacing(2)
        layout.setContentsMargins(0, 0, 0, 0)
        for i in range(fields):
            self._add_item(i)

    def _add_item(self, index):
        _ledit = _NumberValueEdit()
        _ledit.index = index
        _ledit.value_changed.connect(
            lambda: self._on_value_change(_ledit.get_value(), _ledit.index)
        )

        self.layout().addWidget(_ledit)
        self._value.append(0.0)
        self._items.append(_ledit)

    def _on_value_change(self, value=None, index=None):
        if self._can_emit:
            if index is not None:
                self._value = list(self._value)
                self._value[index] = value
            self.value_changed.emit(self.get_name(), self._value)

    def _update_items(self):
        if not isinstance(self._value, (list, tuple)):
            raise TypeError(
                'Value "{}" must be either list or tuple.'.format(self._value)
            )
        for index, value in enumerate(self._value):
            if (index + 1) > len(self._items):
                continue
            if self._items[index].get_value() != value:
                self._items[index].set_value(value)

    def set_data_type(self, data_type):
        """
        Sets the input line edit fields to either display in float or int.

        Args:
            data_type(int or float): int or float data type object.
        """
        for item in self._items:
            item.set_data_type(data_type)

    def set_steps(self, steps):
        """
        Sets the step items in the MMB context menu.

        Args:
            steps (list[int] or list[float]): list of ints or floats.
        """
        for item in self._items:
            item.set_steps(steps)

    def set_min(self, value):
        """
        Set the minimum range for the input fields.

        Args:
            value (int or float): minimum range value.
        """
        for item in self._items:
            item.set_min(value)

    def set_max(self, value):
        """
        Set the maximum range for the input fields.

        Args:
            value (int or float): maximum range value.
        """
        for item in self._items:
            item.set_max(value)

    def get_value(self):
        return self._value

    def set_value(self, value=None):
        if value != self.get_value():
            self._value = value
            self._can_emit = False
            self._update_items()
            self._can_emit = True
            self._on_value_change()


class PropVector2(_PropVector):
    """
    Displays a node property as a "Vector2" widget in the PropertiesBin
    widget.

    Useful for display X,Y data.
    """

    def __init__(self, parent=None):
        super(PropVector2, self).__init__(parent, 2)


class PropVector3(_PropVector):
    """
    Displays a node property as a "Vector3" widget in the PropertiesBin
    widget.

    Useful for displaying x,y,z data.
    """

    def __init__(self, parent=None):
        super(PropVector3, self).__init__(parent, 3)


class PropVector4(_PropVector):
    """
    Displays a node property as a "Vector4"  widget in the PropertiesBin
    widget.

    Useful for display r,g,b,a data.
    """

    def __init__(self, parent=None):
        super(PropVector4, self).__init__(parent, 4)



from caqtus.types.data import DataLabel
from ._proxy import CameraProxy
from .._controller import DeviceController


class CameraController(DeviceController):
    async def run_shot(
        self,
        camera: CameraProxy,
        /,
        timeout: float,
        picture_names: list[str],
        exposures: list[float],
        *args,
        **kwargs,
    ) -> None:
        if len(picture_names) != len(exposures):
            raise ValueError(
                f"Number of picture names ({len(picture_names)}) must be equal to the "
                f"number of exposures ({len(exposures)})"
            )
        await camera.update_parameters(timeout=timeout, *args, **kwargs)
        async with camera.acquire(exposures) as pictures:
            await self.wait_all_devices_ready()
            current_picture = 0
            async for picture in pictures:
                picture_name = picture_names[current_picture]
                label = DataLabel(rf"{self.device_name}\{picture_name}")
                self.signal_data_acquired(
                    label,
                    picture,
                )
                current_picture += 1



from typing import Optional, Callable, Any, TypeVar

import cattrs.strategies
from attr import AttrsInstance
from cattrs import BaseConverter
from cattrs.strategies import (
    include_subclasses as _include_subclasses,
)

from .converters import converters

_C = TypeVar("_C", bound=AttrsInstance)


def include_type(tag_name: str = "class") -> Callable[[Any, BaseConverter], Any]:
    def _include_type(union: Any, converter: BaseConverter) -> Any:
        cattrs.strategies.configure_tagged_union(converter, union, tag_name=tag_name)

    return _include_type


def include_subclasses(
    parent_class: type[_C],
    subclasses: Optional[tuple[type[_C], ...]] = None,
    union_strategy: Optional[Callable[[Any, BaseConverter], Any]] = None,
) -> None:
    for converter in converters.values():
        _include_subclasses(
            parent_class,
            converter,
            subclasses=subclasses,
            union_strategy=union_strategy,
        )


def configure_tagged_union(union: Any, tag_name: str = "type") -> None:
    for converter in converters.values():
        cattrs.strategies.configure_tagged_union(union, converter, tag_name=tag_name)



from __future__ import annotations

import abc
from collections.abc import Mapping
from typing import Any, Union, assert_type, TypeAlias

import attrs

import caqtus.formatter as fmt
from caqtus.types.expression import Expression
from caqtus.types.parameter import Parameters
from caqtus.types.recoverable_exceptions import InvalidTypeError
from caqtus.types.units import is_scalar_quantity, Quantity, dimensionless
from caqtus.types.units.base import BaseUnit
from caqtus.types.variable_name import DottedVariableName


@attrs.define
class Transformation(abc.ABC):
    """Defines a transformation that can be applied to produce an output value."""

    @abc.abstractmethod
    def evaluate(self, variables: Mapping[DottedVariableName, Any]) -> OutputValue:
        """Evaluates the transformation using the given variables.

        If the value returned is a quantity, it must be expressed in base units.
        """

        raise NotImplementedError


type OutputValue = float | int | bool | Quantity[float, BaseUnit]
"""A value that can be used to compute the output of a device.

If the value is a quantity, it must be a scalar and expressed in base units.
"""

EvaluableOutput: TypeAlias = Union[Expression, Transformation]
"""Defines an operation that can be evaluated to an output value.

Evaluable object can be used in the :func:`evaluate` function.
"""


def evaluate(input_: EvaluableOutput, variables: Parameters) -> OutputValue:
    """Evaluates the input and returns the result as a parameter.

    If the evaluated input is a quantity, it is converted to its base units.
    If the output is dimensionless, only the magnitude is returned and the unit is
    stripped.
    """

    if isinstance(input_, Transformation):
        result = input_.evaluate(variables)
    elif isinstance(input_, Expression):
        evaluated = input_.evaluate(variables)

        if isinstance(evaluated, (float, int, bool)):
            result = evaluated
        elif is_scalar_quantity(evaluated):
            result = evaluated.to_base_units()
        else:
            raise InvalidTypeError(
                f"{fmt.expression(input_)} does not evaluate to a parameter, "
                f"got {fmt.type_(type(evaluated))}.",
            )
    assert_type(result, OutputValue)
    if isinstance(result, Quantity) and result.units == dimensionless:
        return result.magnitude
    return result


evaluable_output_validator = attrs.validators.instance_of((Expression, Transformation))



from caqtus.types.recoverable_exceptions import EvaluationError


class UndefinedParameterError(EvaluationError):
    """Indicates that a parameter was not defined in an expression."""


class InvalidOperationError(EvaluationError):
    """Indicates that an invalid operation was attempted."""


class UndefinedUnitError(EvaluationError):
    """Indicates that a unit was not defined in an expression."""



import functools
from typing import assert_never

import numpy as np

import caqtus.formatter as fmt
import caqtus_parsing.nodes as nodes
from caqtus.types.expression import Expression
from caqtus.types.parameter import Parameters
from caqtus.types.recoverable_exceptions import EvaluationError, InvalidValueError
from caqtus.types.units import dimensionless, InvalidDimensionalityError
from caqtus_parsing import parse, InvalidSyntaxError
from ._analog_expression import evaluate_analog_ast
from ._is_time_dependent import is_time_dependent
from .._evaluate_scalar_expression import (
    evaluate_bool_expression,
    evaluate_float_expression,
)
from ...timed_instructions import (
    TimedInstruction,
    Pattern,
    Concatenated,
    concatenate,
    Repeated,
    Ramp,
)
from ...timing import Time, number_ticks

type DigitalInstruction = TimedInstruction[np.bool]


def evaluate_time_dependent_digital_expression(
    expression: Expression, parameters: Parameters, t1: Time, t2: Time, timestep: Time
) -> DigitalInstruction:
    """Evaluate a time-dependent digital expression.

    Args:
        expression: The expression to evaluate.
        parameters: The parameters to use in the evaluation.
        t1: The start time of the evaluation.
        t2: The end time of the evaluation.
        timestep: The time step of the evaluation.

    Returns:
        The result of the evaluation.

    Raises:
        EvaluationError: if an error occurred during evaluation, with the reason for the
            error as the exception cause.
    """

    try:
        ast = parse(str(expression))
        return evaluate_digital_expression(ast, parameters, t1, t2, timestep)
    except (EvaluationError, InvalidSyntaxError) as error:
        raise EvaluationError(
            f"Could not evaluate {fmt.expression(expression)}."
        ) from error


def evaluate_digital_expression(
    expression: nodes.Expression,
    parameters: Parameters,
    t1: Time,
    t2: Time,
    timestep: Time,
) -> DigitalInstruction:
    if not is_time_dependent(expression):
        value = evaluate_bool_expression(expression, parameters)
        length = number_ticks(t1, t2, timestep)
        return Pattern([value]) * length

    match expression:
        case int() | float() | nodes.Quantity():
            raise AssertionError(
                "This should never happen, because at this point, the expression "
                "is known to be time-dependent."
            )
        case nodes.Variable(name=name):
            assert name == "t"
            raise InvalidOperationError(
                f"{fmt.expression(expression)} is not a valid digital expression."
            )
        case (
            nodes.Add()
            | nodes.Subtract()
            | nodes.Multiply()
            | nodes.Divide()
            | nodes.Power()
            | nodes.Plus()
            | nodes.Minus()
        ):
            raise InvalidOperationError(
                f"{fmt.expression(expression)} is not a valid digital expression."
            )
        case nodes.Call():
            return evaluate_call(expression, parameters, t1, t2, timestep)
        case _:
            assert_never(expression)


def evaluate_call(
    call: nodes.Call,
    parameters: Parameters,
    t1: Time,
    t2: Time,
    timestep: Time,
) -> DigitalInstruction:
    if call.function == "square_wave":
        if len(call.args) == 0:
            raise InvalidOperationError(
                f"Function {call.function} requires at least 1 argument, got 0."
            )
        if len(call.args) == 1:
            x_expression = call.args[0]
            duty_cycle = 0.5
        elif len(call.args) == 2:
            x_expression = call.args[0]
            duty_cycle_expression = call.args[1]
            duty_cycle = evaluate_float_expression(duty_cycle_expression, parameters)
            if not 0 <= duty_cycle <= 1:
                raise InvalidValueError(
                    f"Duty cycle {fmt.expression(duty_cycle_expression)} in "
                    f"'square_wave' must be between 0 and 1, got {duty_cycle}."
                )
        else:
            raise InvalidOperationError(
                f"Function {call.function} takes at most 2 arguments, got "
                f"{len(call.args)}."
            )
        x_instr = evaluate_analog_ast(x_expression, parameters, t1, t2, timestep)
        if x_instr.units != dimensionless:
            raise InvalidDimensionalityError(
                f"{fmt.expression(x_expression)} in 'square_wave' must be "
                f"dimensionless, got {x_instr.units}."
            )
        return square_wave(x_instr.magnitudes, duty_cycle)
    else:
        raise InvalidOperationError(
            f"Function {call.function} is not supported in digital expressions."
        )


@functools.singledispatch
def square_wave(x_instr, duty_cycle: float) -> DigitalInstruction:
    raise NotImplementedError(f"Square wave of {type(x_instr)} is not supported.")


@square_wave.register(Pattern)
def square_wave_pattern(
    x_instr: Pattern[np.float64], duty_cycle: float
) -> DigitalInstruction:
    x_values = x_instr.array
    y_values = np.where(x_values % 1 < duty_cycle, True, False)

    return Pattern.create_without_copy(y_values)


@square_wave.register(Concatenated)
def square_wave_concatenated(
    x_instr: Concatenated[np.float64], duty_cycle: float
) -> DigitalInstruction:
    return concatenate(
        *(square_wave(instr, duty_cycle) for instr in x_instr.instructions)
    )


@square_wave.register(Repeated)
def square_wave_repeated(
    x_instr: Repeated[np.float64], duty_cycle: float
) -> DigitalInstruction:
    return x_instr.repetitions * square_wave(x_instr.instruction, duty_cycle)


@square_wave.register(Ramp)
def square_wave_ramp(
    x_instr: Ramp[np.float64], duty_cycle: float
) -> DigitalInstruction:
    # This implementation of square_wave is not totally correct, because it
    # assumes that the resulting values are repetitions of a single pulse.
    # This is not exactly true, when the bounds of the pulse don't always align
    # on the same clock ticks.
    # I am not sure how to handle the more general case.

    slope = (x_instr.stop - x_instr.start) / len(x_instr)
    period = 1 / slope  # period is in clock ticks

    if period < 2:
        raise InvalidValueError(
            "Period of ramp in 'square_wave' must be at least two clock ticks"
        )

    rounded_period = round(period)
    high_duration = round(duty_cycle * rounded_period)
    low_duration = rounded_period - high_duration
    if high_duration == 0 and duty_cycle != 0:
        raise InvalidValueError(
            "High duration of square wave in smaller than one clock tick"
        )
    if low_duration == 0 and duty_cycle != 1:
        raise InvalidValueError(
            "Low duration of square wave in smaller than one clock tick"
        )

    pulse = Pattern([True]) * high_duration + Pattern([False]) * low_duration
    initial_x = x_instr.start % 1

    cut = round(initial_x * period)
    pulse = pulse[cut:] + pulse[:cut]

    repetitions, remainder = divmod(len(x_instr), rounded_period)
    return pulse * repetitions + pulse[:remainder]


class InvalidOperationError(EvaluationError):
    """Raised when an invalid operation is attempted."""

    pass
